"""Markov clustering tests."""

import numpy as np
from ragraph.analysis import cluster
from ragraph.analysis.cluster._markov import MarkovFlow
from ragraph.graph import Graph


def test_markov_flow():
    matrix = np.array([[0.0, 0.0], [1.0, 0.0]])

    mf = MarkovFlow(matrix, 2.0, False)
    assert mf.dim == 2
    assert mf.sink_matrix.tolist() == [
        [0.0, 0.0, 0.0],
        [0.5, 0.0, 0.0],
        [0.5, 1.0, 0.0],
    ]
    assert mf.sensitivity_matrix.tolist() == [
        [1.0, 0.0, 0.0],
        [0.5, 1.0, 0.0],
        [1.0, 1.0, 1.0],
    ]
    assert mf.in_vector.flatten().tolist() == [1.0, 1.0, 0.0]
    assert mf.flow_vector.flatten().tolist() == [1.0, 1.5, 2.0]
    assert mf.flow_matrix.tolist() == [
        [0.0, 0.0, 0.0],
        [0.5, 0.0, 0.0],
        [0.5, 1.5, 0.0],
    ]


def test_markov_defaults(graph: Graph):
    notinplace, _ = cluster.markov(graph, inplace=False)
    cluster.markov(graph)
    assert graph.get_hierarchy_dict() == {
        "f": {},
        "node.node0": {"a": {}, "b": {}, "c": {}, "d": {}, "e": {}},
    }
    assert notinplace.get_hierarchy_dict() == graph.get_hierarchy_dict()


def test_hierarchical_markov_defaults(graph: Graph):
    ref = {
        "f": {},
        "node.node2": {
            "node.node0": {"a": {}, "b": {}},
            "node.node1": {"c": {}, "d": {}, "e": {}},
        },
    }
    notinplace, _ = cluster.hierarchical_markov(graph, beta=4.0, mu=6.0, inplace=False)
    assert notinplace.get_hierarchy_dict() == ref
    cluster.hierarchical_markov(graph, beta=4.0, mu=6.0)
    value = graph.get_hierarchy_dict()
    assert value == ref


def test_single_node_edge(graph: Graph):
    assert cluster.markov(graph, leafs=["a"], names=True)[1] == ["a"]
    assert cluster.hierarchical_markov(graph, leafs=["a"], names=True)[1] == ["a"]


def test_cluster_markov_cases(case: Graph):
    cluster.markov(case)


def test_cluster_hierarchical_markov_cases(case: Graph):
    cluster.hierarchical_markov(case)


def test_hierarchical_markov_paper_case(paper: Graph):
    graph, roots = cluster.hierarchical_markov(paper, alpha=2, beta=3.0, mu=3.0, inplace=False)
    assert graph



"""Tests for Tarjan's clustering algorithms module."""

from typing import List, Set

from ragraph.analysis import cluster
from ragraph.analysis.cluster._tarjan import (
    tarjans_scc,
    tarjans_scc_algorithm,
)
from ragraph.graph import Graph
from ragraph.node import Node


def test_tarjans_scc_algorithm(graph: Graph):
    sccs = tarjans_scc_algorithm(graph, graph.leafs, True)
    sets: List[Set[Node]] = [set(scc) for scc in sccs]

    assert sets == [
        set(graph.nodes[:-1]),
        set(graph.nodes[-1:]),
    ], "Sets should match predicted output ({a,b,c,d,e}, {f})."
    assert sccs[1] == [graph.node_dict["f"]], "Doublecheck f should be a single SCC."


def test_tarjans_scc_clustering(graph: Graph):
    notinplace, _ = tarjans_scc(graph, root=None, leafs=graph.leafs, edge_weights=[], inplace=False)
    tarjans_scc(graph, root=None, leafs=graph.leafs, edge_weights=[], inplace=True)
    assert graph.get_hierarchy_dict() == {
        "f": {},
        "node.node0": {"a": {}, "b": {}, "c": {}, "d": {}, "e": {}},
    }
    assert notinplace.get_hierarchy_dict() == graph.get_hierarchy_dict()


def test_tarjans_scc_cases(case: Graph):
    cluster.tarjans_scc(case)



"""Tests for edge.py"""

import pytest
from ragraph.edge import Edge
from ragraph.node import Node


@pytest.fixture
def rich_edge():
    source = Node("source")
    target = Node("target")
    return Edge(
        source,
        target,
        name="rich edge",
        kind="rich",
        labels=["money", "mucho"],
        weights=dict(dollars=1, euros=0.5),
        annotations=dict(foo="bar"),
    )


def test_edge_properties(rich_edge):
    assert rich_edge.source.name == "source"
    assert rich_edge.target.name == "target"
    assert rich_edge.kind == "rich"
    assert rich_edge.labels == ["money", "mucho"]
    assert rich_edge.weights == dict(dollars=1, euros=0.5)
    assert rich_edge.weight == 1.5
    assert rich_edge.annotations.foo == "bar"


def test_edge_instantiations():
    assert Edge("test", "test", kind=None).kind == "default", "Default edge kind should be 'edge'."

    assert (
        Edge("test", "test", annotations=dict(a=1)).annotations.a == 1
    ), "Edge annotations should accept a dict as argument."


def test_edge_representation(rich_edge):
    assert str(rich_edge) == "Edge(source='source', target='target', name='rich edge')"
    non_id = repr(rich_edge).split(", uuid=")[0]
    assert non_id == (
        "<ragraph.edge.Edge(source=Node(name='source'), target=Node(name='target'), "
        + "name='rich edge', kind='rich', labels=['money', 'mucho'], "
        + "weights={'dollars': 1, 'euros': 0.5}, "
        + "annotations=Annotations({'foo': 'bar'})"
    )



"""Tests for Markov sequencing."""

from ragraph import datasets
from ragraph.analysis import sequence
from ragraph.analysis.sequence.utils import markov_decision
from ragraph.io.matrix import from_matrix


def test_markov_sequencing_chain(chain_graph):
    assert chain_graph is not None
    _, seq = sequence.markov(chain_graph, inf=1.0, dep=1.0, mu=2.0, names=True)
    assert seq == ["4", "3", "2", "1", "0"]

    single = [chain_graph.nodes[0]]
    assert sequence.markov(chain_graph, nodes=single)[1] == single


def test_markov_sequencing_ucav():
    casename = "ucav"
    case = datasets.get(casename)
    _, seq = sequence.markov(
        case,
        nodes=[n.name for n in case.leafs],
        inf=1.0,
        dep=1.0,
        mu=1.5,
        edge_weights=["binary"],
        names=True,
    )
    ref = [
        "Prepare UCAV Preliminary DR&O",
        "Develop Structural Design Conditions",
        "Perform Aerodynamics Analyses & Evaluation",
        "Perform Weights & Inertias Analyses & Evaluation",
        "Create UCAV Preliminary Design Configuration",
        "Prepare Structural Geometry & Notes for FEM",
        "Prepare & Distribute Surfaced Models & Internal Drawings",
        "Perform S&C Analyses & Evaluation",
        "Establish Internal Load Distributions",
        "Preliminary Manufacturing Planning & Analyses",
        "Develop Balanced Freebody Diagrams & External Loads",
        "Evaluate Structural Strength, Stiffness, & Life",
        "Create Initial Structural Geometry",
        "Prepare UCAV Proposal",
    ]
    assert seq == ref


def test_markov_sequencing_cases(case):
    assert case is not None
    sequence.markov(case, nodes=[n.name for n in case.leafs], inf=2.0, dep=1.0, mu=2.0)


def test_markov_sequence_bus_behavior():
    graph = from_matrix(
        [
            [0, 5, 5, 5],
            [1, 0, 0, 0],
            [1, 0, 0, 0],
            [1, 0, 0, 0],
        ],
        rows="abcd",
    )

    _, seq = sequence.markov(graph, names=True)

    assert seq[-1] == "a"

    seq = sequence.branchsort(graph, algo=sequence.scc_tearing, names=True)[2]

    assert seq[-1] == "a"

    nodes = graph.leafs
    assert nodes[markov_decision(graph, nodes)] != "a"



"""# GRIP format support"""

from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional, Union
from uuid import UUID, uuid4

from lxml import etree

from ragraph import logger
from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node

REPORT_CATEGORIES = [
    "Documenten",
    "Functies",
    "Objecttypen",
    "Organisaties",
    "Systeemeisen",
    "Scope",
    "Raakvlakken",
    "Begrippen",
    "StructuurItems",
    "Managementeisen",
    "Klanteisen",
    "Objecten",
    "Keuzevragen",
    "Producteisen",
    "Processtructuur",
    "RIsicos",
    "Procesrisicos",
    "Onderhoudsproducten",
    "Proceseisen",
    "Onderhoudsactiviteiten",
]

REPORT_CATEGORY_PARAMS = {
    "Documenten": ["ID", "User"],
    "Functies": ["ID", "User"],
    "Objecttypen": ["ID", "User"],
    "Organisaties": ["ID", "User", "Type_uitwisseling"],
    "Systeemeisen": ["ID", "User", "Type_uitwisseling"],
    "Scope": [
        "ID",
        "User",
        "Type_uitwisseling",
        "Eisteksten_aanbieden_als",
        "HerkomstID_vullen_met",
        "Versie_specificatie",
    ],
    "Raakvlakken": ["ID", "User", "Type_uitwisseling"],
    "Begrippen": ["ID", "User", "Type_uitwisseling"],
    "StructuurItems": ["ID", "User", "Type_uitwisseling"],
    "Managementeisen": ["ID", "User", "Type_uitwisseling"],
    "Klanteisen": ["ID", "User", "Type_uitwisseling"],
    "Objecten": ["ID", "User", "Type_uitwisseling"],
    "Keuzevragen": ["ID", "User", "Type_uitwisseling"],
    "Producteisen": ["ID", "User", "Type_uitwisseling"],
    "Processtructuur": ["ID", "User", "Type_uitwisseling"],
    "RIsicos": ["ID", "User", "Type_uitwisseling"],
    "Procesrisicos": ["ID", "User", "Type_uitwisseling"],
    "Onderhoudsproducten": ["ID", "User", "Type_uitwisseling"],
    "Proceseisen": ["ID", "User", "Type_uitwisseling"],
    "Onderhoudsactiviteiten": ["ID", "User", "Type_uitwisseling"],
}


OBJECT_KIND = "object"
OBJECTTYPE_KIND = "objecttype"
FUNCTIE_KIND = "functie"
SYSTEEMEIS_KIND = "systeemeis"
SCOPE_KIND = "scope"
RAAKVLAK_KIND = "raakvlak"
ACTIVITY_KIND = "activity"
PARAMETER_KIND = "param"
INHERITANCE_KIND = "inheritance"


HIERARCHY_MODE_ROOTREFS = {
    "bottom-up": "e0021f8f-114b-ee11-b6a6-001dd8d7027d",
    "top-down": "855342d6-3895-e211-81b6-001d09fa6b1e",
}


def from_grip(path: Union[str, Path], hierarchy_mode: str = "top-down") -> Graph:
    """Decode GRIP XML file, string, or element into a Graph.

    Arguments:
        path: GRIP XML file path.
        hierarchy_mode: One of "bottom-up" or "top-down".
            Defines how the hierarchy relations are stored.

    Returns:
        Graph object.
    """

    tree = etree.parse(str(path))
    root = tree.getroot()

    graph = Graph(
        uuid=root.attrib.get("WorkspaceID"),
        name=root.attrib.get("WorkspaceName"),
        annotations=dict(root.attrib),
    )
    parse_params(graph, root)

    parse_collection(graph, root, "Objecten", "Object", OBJECT_KIND)
    parse_objectenboom(graph, root, hierarchy_mode)

    parse_collection(graph, root, "Objecttypen", "Objecttype", OBJECTTYPE_KIND)
    parse_objecttypenboom(graph, root)

    parse_collection(graph, root, "Functies", "Functie", FUNCTIE_KIND)
    parse_collection(graph, root, "Systeemeisen", "Systeemeis", SYSTEEMEIS_KIND)
    parse_collection(graph, root, "Onderhoudsactiviteiten", "Onderhoudsactiviteit", ACTIVITY_KIND)
    parse_collection(graph, root, "Scope", "Scope", SCOPE_KIND)
    parse_collection(graph, root, "Raakvlakken", "Raakvlak", RAAKVLAK_KIND)

    parse_systeemeis_edges(graph, root)
    parse_object_edges(graph, root)
    parse_scope_edges(graph, root)
    parse_raakvlak_edges(graph, root)
    parse_activity_edges(graph, root)

    # Enrich graph.
    add_raakvlak_annotations(graph)
    add_labels(graph)

    return graph


def parse_params(graph: Graph, root: etree.Element):
    """Decode parameter section of GRIP XML file.

    Arguments:
       graph: Graph object to add nodes to.
       root: Root of XML file.
    """
    scope = root.find("Scope")
    for param in scope.find("RelaticsParameters").iterfind("RelaticsParameter"):
        graph.add_node(Node(kind=PARAMETER_KIND, annotations=param.attrib))


def parse_eisteksten(el: etree.Element, annotations: Dict):
    """Parse eisteksten and store within annotations.

    Arguments:
       el: Systeemeis to be parsed
       annotations: Dictionary in which information must be stored
    """
    if el.find("CI_EistekstDefinitief") is not None:
        annotations.update(
            {
                "CI_EistekstDefinitief": {
                    "R1Sequence": el.find("CI_EistekstDefinitief").attrib["R1Sequence"],
                    "R2Sequence": el.find("CI_EistekstDefinitief").attrib["R2Sequence"],
                    "RootRef": el.find("CI_EistekstDefinitief").attrib["RootRef"],
                    "Type": el.find("CI_EistekstDefinitief").attrib["Type"],
                    "Eistekst": el.find("CI_EistekstDefinitief").find("Eistekst").attrib,
                }
            }
        )

    if el.find("CI_EistekstOrigineel") is not None:
        annotations.update(
            {
                "CI_EistekstOrigineel": {
                    "R1Sequence": el.find("CI_EistekstOrigineel").attrib["R1Sequence"],
                    "R2Sequence": el.find("CI_EistekstOrigineel").attrib["R2Sequence"],
                    "RootRef": el.find("CI_EistekstOrigineel").attrib["RootRef"],
                    "Type": el.find("CI_EistekstOrigineel").attrib["Type"],
                    "EistekstOrigineel": el.find("CI_EistekstOrigineel")
                    .find("EistekstOrigineel")
                    .attrib,
                }
            }
        )


def parse_CI_MEEisObjects(el: etree.Element, annotations: Dict):
    """Parse CI_MEEisObjects and store within annotations.

    Arguments:
       el: Systeemeis to be parsed
       annotations: Dictionary in which information must be stored
    """
    annotations["CI_MEEisObjects"] = []

    for MObj in el.iterfind("CI_MEEisObject"):
        annotations["CI_MEEisObjects"].append(
            {
                "CI_MEEisObject": {
                    "R1Sequence": MObj.attrib["R1Sequence"],
                    "R2Sequence": MObj.attrib["R2Sequence"],
                    "RootRef": MObj.attrib["RootRef"],
                    "Type": MObj.attrib["Type"],
                    "EisObject": {
                        "Name": MObj.find("EisObject").attrib["Name"],
                        "ConfigurationOfRef": MObj.find("EisObject").attrib["ConfigurationOfRef"],
                        "Type": MObj.find("EisObject").attrib["Type"],
                        "GUID": MObj.find("EisObject").attrib["GUID"],
                        "SI_Object": {
                            "R1Sequence": MObj.find("EisObject")
                            .find("SI_Object")
                            .attrib["R1Sequence"],
                            "R2Sequence": MObj.find("EisObject")
                            .find("SI_Object")
                            .attrib["R2Sequence"],
                            "RootRef": MObj.find("EisObject").find("SI_Object").attrib["RootRef"],
                            "Type": MObj.find("EisObject").find("SI_Object").attrib["Type"],
                            "Object": MObj.find("EisObject")
                            .find("SI_Object")
                            .find("Object")
                            .attrib,
                        },
                    },
                }
            }
        )


def parse_CI_MEEisObjecttypen(el: etree.Element, annotations: Dict):
    """Parse CI_MEEisObjectypen and store within annotations.

    Arguments:
       el: Systeemeis to be parsed
       annotations: Dictionary in which information must be stored
    """
    annotations["CI_MEEisObjecttypen"] = []

    for MObjtype in el.iterfind("CI_MEEisObjecttype"):
        annotations["CI_MEEisObjecttypen"].append(
            {
                "CI_MEEisObjecttype": {
                    "R1Sequence": MObjtype.attrib["R1Sequence"],
                    "R2Sequence": MObjtype.attrib["R2Sequence"],
                    "RootRef": MObjtype.attrib["RootRef"],
                    "Type": MObjtype.attrib["Type"],
                    "EisObjecttype": {
                        "Name": MObjtype.find("EisObjecttype").attrib["Name"],
                        "ConfigurationOfRef": MObjtype.find("EisObjecttype").attrib[
                            "ConfigurationOfRef"
                        ],
                        "Type": MObjtype.find("EisObjecttype").attrib["Type"],
                        "GUID": MObjtype.find("EisObjecttype").attrib["GUID"],
                        "SI_Objecttype": {
                            "R1Sequence": MObjtype.find("EisObjecttype")
                            .find("SI_Objecttype")
                            .attrib["R1Sequence"],
                            "R2Sequence": MObjtype.find("EisObjecttype")
                            .find("SI_Objecttype")
                            .attrib["R2Sequence"],
                            "RootRef": MObjtype.find("EisObjecttype")
                            .find("SI_Objecttype")
                            .attrib["RootRef"],
                            "Type": MObjtype.find("EisObjecttype")
                            .find("SI_Objecttype")
                            .attrib["Type"],
                            "Objecttype": MObjtype.find("EisObjecttype")
                            .find("SI_Objecttype")
                            .find("Objecttype")
                            .attrib,
                        },
                    },
                }
            }
        )

def parse_SI_Eistype(el: etree.Element, annotations: Dict):
    """Parse information attached to an SI_Eistype."""
    st = el.find("SI_Eistype")

    annotations["SI_Eistype"] = {
        "R1Sequence": st.attrib["R1Sequence"],
        "R2Sequence": st.attrib["R2Sequence"],
        "RootRef": st.attrib["RootRef"],
        "Type": st.attrib["Type"],
        "Eistype": {
            "Name": st.find("Eistype").attrib["Name"],
            "ConfigurationOfRef": st.find("Eistype").attrib["ConfigurationOfRef"],
            "GUID": st.find("Eistype").attrib["GUID"]
        }
    }


def parse_SI_Periode(el: etree.Element, annotations: Dict):
    """Parse information attached to an SI_Periode."""
    st = el.find("SI_Periode")

    annotations["SI_Periode"] = {
        "R1Sequence": st.attrib["R1Sequence"],
        "R2Sequence": st.attrib["R2Sequence"],
        "RootRef": st.attrib["RootRef"],
        "Type": st.attrib["Type"],
        "Periode": {
            "Name": st.find("Periode").attrib["Name"],
            "ConfigurationOfRef": st.find("Periode").attrib["ConfigurationOfRef"],
            "GUID": st.find("Periode").attrib["GUID"]
        }
    }


def parse_SI_Systeemeisgroepering(el: etree.Element, annotations: Dict):
    """Parse information attached to an SI_Systeemeisgroepering."""
    annotations["SI_Systeemeisgroepering"] = []

    for st in el.iterfind("SI_Systeemeisgroepering"):

        annotations["SI_Systeemeisgroepering"].append({
            "R1Sequence": st.attrib["R1Sequence"],
            "R2Sequence": st.attrib["R2Sequence"],
            "RootRef": st.attrib["RootRef"],
            "Type": st.attrib["Type"],
            "Systeemeisgroepering": {
                "Name": st.find("Systeemeisgroepering").attrib["Name"],
                "ConfigurationOfRef": st.find("Systeemeisgroepering").attrib["ConfigurationOfRef"],
                "GUID": st.find("Systeemeisgroepering").attrib["GUID"]
            }
        })


def parse_CI_Verificatievoorschrift(el: etree.Element, annotations: Dict):
    """Parse information attached to an CI verificationvoorschrift"""
    
    annotations["CI_Verificatievoorschrift"] = []

    for cvv in el.iterfind("CI_Verificatievoorschrift"):
        info = {
            "R1Sequence": cvv.attrib["R1Sequence"],
            "R2Sequence": cvv.attrib["R2Sequence"],
            "RootRef": cvv.attrib["RootRef"],
            "Type": cvv.attrib["Type"],
        }

        parse_verificatie(cvv, info)

        annotations["CI_Verificatievoorschrift"].append(info)


def parse_verificatie(el: etree.Element, annotations: Dict):
    """Parse information attached to Verificatie."""
    v = el.find("Verificatie")

    annotations["Verificatie"] = {
        "Name": v.attrib["Name"],
        "ConfigurationOfRef": v.attrib["ConfigurationOfRef"],
        "GUID": v.attrib["GUID"],
        "Criterium": {
            "RootRef": v.find("Criterium").attrib["RootRef"],
            "Type": v.find("Criterium").attrib["Type"]
        }
    }

    parse_SI_Methode(v, annotations["Verificatie"])
    parse_SI_Fase(v, annotations["Verificatie"])
    parse_CI_Toelichting(v, annotations["Verificatie"])


def parse_SI_Methode(el: etree.Element, annotations: Dict):
    """Parse information attached to an SI_Methode."""
    st = el.find("SI_Methode")

    annotations["SI_Methode"] = {
        "R1Sequence": st.attrib["R1Sequence"],
        "R2Sequence": st.attrib["R2Sequence"],
        "RootRef": st.attrib["RootRef"],
        "Type": st.attrib["Type"],
        "Methode": {
            "Name": st.find("Methode").attrib["Name"],
            "ConfigurationOfRef": st.find("Methode").attrib["ConfigurationOfRef"],
            "GUID": st.find("Methode").attrib["GUID"]
        }
    }


def parse_SI_Fase(el: etree.Element, annotations: Dict):
    """Parse information attached to an SI_Fase."""
    st = el.find("SI_Fase")

    annotations["SI_Fase"] = {
        "R1Sequence": st.attrib["R1Sequence"],
        "R2Sequence": st.attrib["R2Sequence"],
        "RootRef": st.attrib["RootRef"],
        "Type": st.attrib["Type"],
        "Fase": {
            "Name": st.find("Fase").attrib["Name"],
            "ConfigurationOfRef": st.find("Fase").attrib["ConfigurationOfRef"],
            "GUID": st.find("Fase").attrib["GUID"]
        }
    }


def parse_CI_Toelichting(el: etree.Element, annotations: Dict):
    """Parse information attached to an CI_Toelichting."""
    st = el.find("CI_Toelichting")

    annotations["CI_Toelichting"] = {
        "R1Sequence": st.attrib["R1Sequence"],
        "R2Sequence": st.attrib["R2Sequence"],
        "RootRef": st.attrib["RootRef"],
        "Type": st.attrib["Type"],
        "Toelichting": {
            "Name": st.find("Toelichting").attrib["Name"],
            "Description": st.find("Toelichting").attrib.get("Description", ""),
            "ConfigurationOfRef": st.find("Toelichting").attrib["ConfigurationOfRef"],
            "GUID": st.find("Toelichting").attrib["GUID"]
        }
    }

def parse_CI_MEOnderhoudsactiviteitObject(el: etree.Element, annotations: Dict):
    """Parse information for Activity object"""
    annotations["CI_MEOnderhoudsactiviteitObjecten"] = []

    for MEOa in el.iterfind("CI_MEOnderhoudsactiviteitObject"):
        annotations["CI_MEOnderhoudsactiviteitObjecten"].append(
            {
                "CI_MEOnderhoudsactiviteitObject": {
                    "R1Sequence": MEOa.attrib["R1Sequence"],
                    "R2Sequence": MEOa.attrib["R2Sequence"],
                    "RootRef": MEOa.attrib["RootRef"],
                    "Type": MEOa.attrib["Type"],
                    "MEOnderhoudsactiviteitObject": {
                        "Name": MEOa.find("MEOnderhoudsactiviteitObject").attrib["Name"],
                        "ConfigurationOfRef": MEOa.find("MEOnderhoudsactiviteitObject").attrib[
                            "ConfigurationOfRef"
                        ],
                        "Type": MEOa.find("MEOnderhoudsactiviteitObject").attrib["Type"],
                        "GUID": MEOa.find("MEOnderhoudsactiviteitObject").attrib["GUID"],
                        "SI_Object": {
                            "R1Sequence": MEOa.find("MEOnderhoudsactiviteitObject")
                            .find("SI_Object")
                            .attrib["R1Sequence"],
                            "R2Sequence": MEOa.find("MEOnderhoudsactiviteitObject")
                            .find("SI_Object")
                            .attrib["R2Sequence"],
                            "RootRef": MEOa.find("MEOnderhoudsactiviteitObject")
                            .find("SI_Object")
                            .attrib["RootRef"],
                            "Type": MEOa.find("MEOnderhoudsactiviteitObject")
                            .find("SI_Object")
                            .attrib["Type"],
                            "Object": MEOa.find("MEOnderhoudsactiviteitObject")
                            .find("SI_Object")
                            .find("Object")
                            .attrib,
                        },
                    },
                }
            }
        )

def parse_element(el: etree.Element, item: str) -> Dict:
    """Parsing element and store relevant information within Node annotations

    Arguments:
      el: Element to parse.
      item: item kind being parsed.

    Returns:
      Name string and Annotations dictionary.
    """
    annotations = dict(el.attrib)

    if el.find("ID") is not None:
        annotations.update(dict(ID=el.find("ID").attrib))
        if item == "Onderhoudsactiviteit":
            id1 = el.find("ID").attrib.get("IDOA")
        else:
            id1 = el.find("ID").attrib.get("ID1")
        name = el.attrib.get("Name")
        if name and id1 not in name:
            name = f"{name} | {id1}"
        elif not name:
            name = id1
    else:
        name = el.attrib.get("Name")

    for key in ["Code", "BronID", "Eiscodering", "ExterneCode"]:
        if el.find(key) is None:
            continue
        annotations.update({key: el.find(key).attrib})

    if item == "Systeemeis":
        parse_eisteksten(el, annotations)

        if el.find("CI_MEEisObject") is not None:
            parse_CI_MEEisObjects(el=el, annotations=annotations)

        if el.find("CI_MEEisObjecttype") is not None:
            parse_CI_MEEisObjecttypen(el=el, annotations=annotations)

        if el.find("SI_Eistype") is not None:
            parse_SI_Eistype(el=el, annotations=annotations)

        if el.find("SI_Periode") is not None:
            parse_SI_Periode(el=el, annotations=annotations)

        if el.find("SI_Systeemeisgroepering") is not None:
            parse_SI_Systeemeisgroepering(el=el, annotations=annotations)

        if el.find("CI_Verificatievoorschrift") is not None:
            parse_CI_Verificatievoorschrift(el=el, annotations=annotations)

        if el.find("CI_Toelichting") is not None:
            parse_CI_Toelichting(el=el, annotations=annotations)
                

    # Store information for activities
    if item == "Onderhoudsactiviteit":
        if el.find("CI_MEOnderhoudsactiviteitObject") is not None:
            parse_CI_MEOnderhoudsactiviteitObject(el=el, annotations=annotations)
        if el.find("Aard") is not None:
            annotations["Aard"] = el.find("Aard").attrib
        if el.find("Onderhoudsactiviteittype") is not None:
            annotations["Onderhoudsactiviteittype"] = el.find("Onderhoudsactiviteittype").attrib

    # Parse hierarchy relations for objects
    if el.find("SI_Onderliggend") is not None:
        annotations.update({"RootRefOnderliggend": el.find("SI_Onderliggend").attrib["RootRef"]})

    if el.find("SI_Bovenliggend") is not None:
        annotations.update({"RootRefBovenliggend": el.find("SI_Bovenliggend").attrib["RootRef"]})

    # Parse hiearchy relations for objecttypes.
    if el.find("SI_Onderliggende") is not None:
        annotations.update({"RootRefOnderliggende": el.find("SI_Onderliggende").attrib["RootRef"]})

    # Parse relations to functions.
    if el.find("SI_Functie") is not None:
        annotations.update({"RootRefFunctie": el.find("SI_Functie").attrib["RootRef"]})

    # Parse assignment relations to objectttpes.
    if el.find("SI_Objecttype") is not None:
        annotations.update({"RootRefObjecttype": el.find("SI_Objecttype").attrib["RootRef"]})

    return name, annotations


def parse_collection(graph: Graph, root: etree.Element, collection: str, item: str, kind: str):
    """Decode contents of XML file.

    Arguments:
       graph: Graph object to add nodes to
       root: Root of XML file
       collection: Sub-collection to parse (e.g. objecten or Objecttypen)
       item: item kind to parse.
       kind: kind to assign to created nodes.
    """
    coll = root.find(collection)

    if coll is None:
        "No collection found, return."
        return 

    for el in coll.iterfind(item):
        name, annotations = parse_element(el=el, item=item)
        graph.add_node(
            Node(
                uuid=el.attrib.get("GUID"),
                name=name,
                kind=kind,
                annotations=annotations,
            )
        )


def parse_objectenboom(graph: Graph, root: etree.Element, hierarchy_mode: str = "bottom-up"):
    collection = root.find("Objecten")
    for el in collection.iterfind("Object"):
        if hierarchy_mode == "top-down":
            parent = graph[UUID(el.attrib.get("GUID"))]
            for sub in el.iterfind("SI_Onderliggend"):
                for obj in sub.iterfind("ObjectOnderliggend"):
                    child_id = UUID(obj.attrib.get("GUID"))
                    graph[child_id].parent = parent
        elif hierarchy_mode == "bottom-up":
            child = graph[UUID(el.attrib.get("GUID"))]
            for sub in el.iterfind("SI_Bovenliggend"):
                for obj in sub.iterfind("ObjectBovenliggend"):
                    parent_id = UUID(obj.attrib.get("GUID"))
                    child.parent = graph[parent_id]


def parse_objecttypenboom(graph: Graph, root: etree.Element):
    collection = root.find("Objecttypen")
    for el in collection.iterfind("Objecttype"):
        parent = graph[UUID(el.attrib.get("GUID"))]
        for sub in el.iterfind("SI_Onderliggende"):
            for obj in sub.iterfind("Objecttype"):
                child_id = UUID(obj.attrib.get("GUID"))
                graph.add_edge(Edge(source=graph[child_id], target=parent, kind="inheritance"))


def parse_systeemeis_edges(graph: Graph, root: etree.Element):
    elems = root.find("Systeemeisen").iterfind("Systeemeis")
    for el in elems:
        source = graph[UUID(el.attrib.get("GUID"))]

        for me_eis in el.iterfind("CI_MEEisObject"):
            eis_obj = me_eis.find("EisObject")
            eis_obj.attrib.get("GUID")
            object_id = eis_obj.find("SI_Object").find("Object").attrib.get("GUID")
            target = graph[UUID(object_id)]
            graph.add_edge(Edge(source, target, kind=SYSTEEMEIS_KIND))
            graph.add_edge(Edge(target, source, kind=SYSTEEMEIS_KIND))

        for me_eis in el.iterfind("CI_MEEisObjecttype"):
            eis_obj = me_eis.find("EisObjecttype")
            eis_obj.attrib.get("GUID")
            object_id = eis_obj.find("SI_Objecttype").find("Objecttype").attrib.get("GUID")
            target = graph[UUID(object_id)]
            graph.add_edge(Edge(source, target, kind=SYSTEEMEIS_KIND))
            graph.add_edge(Edge(target, source, kind=SYSTEEMEIS_KIND))

        for sub in el.iterfind("SI_Functie"):
            annotations = dict(RootRef=sub.attrib["RootRef"])
            for functie in sub.iterfind("Functie"):
                functie_id = UUID(functie.attrib.get("GUID"))
                target = graph[functie_id]
                graph.add_edge(Edge(source, target, kind=OBJECT_KIND, annotations=annotations))
                graph.add_edge(Edge(target, source, kind=OBJECT_KIND, annotations=annotations))


def parse_object_edges(graph: Graph, root: etree.Element):
    collection = root.find("Objecten")
    for el in collection.iterfind("Object"):
        source = graph[UUID(el.attrib.get("GUID"))]

        for sub in el.iterfind("SI_Functie"):
            annotations = dict(RootRef=sub.attrib["RootRef"])
            for functie in sub.iterfind("Functie"):
                functie_id = UUID(functie.attrib.get("GUID"))
                target = graph[functie_id]
                graph.add_edge(Edge(source, target, kind=OBJECT_KIND, annotations=annotations))
                graph.add_edge(Edge(target, source, kind=OBJECT_KIND, annotations=annotations))

        for sub in el.iterfind("SI_Objecttype"):
            annotations = dict(RootRef=sub.attrib["RootRef"])
            for objecttype in sub.iterfind("Objecttype"):
                objecttype_id = UUID(objecttype.attrib.get("GUID"))
                target = graph[objecttype_id]
                graph.add_edge(Edge(source, target, kind="inheritance", annotations=annotations))
                graph.add_edge(Edge(target, source, kind="inheritance", annotations=annotations))


def parse_scope_edges(graph: Graph, root: etree.Element):
    elems = root.find("Scope").iterfind("Scope")
    for el in elems:
        source = graph[UUID(el.attrib.get("GUID"))]

        for eis in el.iterfind("SI_Systeemeis"):
            annotations = dict(
                RootRef=eis.attrib.get("RootRef"),
                R1Sequence=eis.attrib.get("R1Sequence"),
                R2Sequence=eis.attrib.get("R2Sequence"),
            )
            eis_id = eis.find("Systeemeis").attrib.get("GUID")
            target = graph[UUID(eis_id)]
            graph.add_edge(Edge(source, target, kind=SCOPE_KIND, annotations=annotations))
            graph.add_edge(Edge(target, source, kind=SCOPE_KIND, annotations=annotations))

        for functie in el.iterfind("SI_Functie"):
            annotations = dict(
                RootRef=functie.attrib.get("RootRef"),
                R1Sequence=functie.attrib.get("R1Sequence"),
                R2Sequence=functie.attrib.get("R2Sequence"),
                RootRefFunctie=functie.attrib.get("RootRef")
            )
            functie_id = functie.find("Functie").attrib.get("GUID")
            target = graph[UUID(functie_id)]
            graph.add_edge(Edge(source, target, kind=SCOPE_KIND, annotations=annotations))
            graph.add_edge(Edge(target, source, kind=SCOPE_KIND, annotations=annotations))

        for raakvlak in el.iterfind("SI_Raakvlak"):
            annotations = dict(
                RootRef=raakvlak.attrib.get("RootRef"),
                R1Sequence=raakvlak.attrib.get("R1Sequence"),
                R2Sequence=raakvlak.attrib.get("R2Sequence"),
            )
            if raakvlak.find("SI_Objecttype") is not None:
                annotations["SI_ObjecttypeRootRef"] = raakvlak.find("SI_Objecttype")["RootRef"]

            raakvlak_id = raakvlak.find("Raakvlak").attrib.get("GUID")

            try:
                target = graph[UUID(raakvlak_id)]
            except KeyError:
                logger.warning(f"UUID {UUID(raakvlak_id)} not found.")

            graph.add_edge(Edge(source, target, kind=SCOPE_KIND, annotations=annotations))
            graph.add_edge(Edge(target, source, kind=SCOPE_KIND, annotations=annotations))

        for obj in el.iterfind("SI_Object"):
            annotations = dict(
                RootRef=obj.attrib.get("RootRef"),
                R1Sequence=obj.attrib.get("R1Sequence"),
                R2Sequence=obj.attrib.get("R2Sequence"),
            )
            obj_id = obj.find("Object").attrib.get("GUID")

            try:
                target = graph[UUID(obj_id)]
            except KeyError:
                logger.warning(f"UUID {UUID(obj_id)} not found.")

            graph.add_edge(Edge(source, target, kind=SCOPE_KIND, annotations=annotations))
            graph.add_edge(Edge(target, source, kind=SCOPE_KIND, annotations=annotations))


def parse_activity_edges(graph: Graph, root: etree.Element):
    if root.find("Onderhoudsactiviteiten") is None:
        "No 'Onderhoudsactiviteiten', return."
        return
    
    elems = root.find("Onderhoudsactiviteiten").iterfind("Onderhoudsactiviteit")
    for el in elems:
        source = graph[UUID(el.attrib.get("GUID"))]

        for ci_obj in el.iterfind("CI_MEOnderhoudsactiviteitObject"):
            me_obj = ci_obj.find("MEOnderhoudsactiviteitObject")
            object_id = me_obj.find("SI_Object").find("Object").attrib.get("GUID")
            target = graph[UUID(object_id)]
            graph.add_edge(Edge(source, target, kind=ACTIVITY_KIND))
            graph.add_edge(Edge(target, source, kind=ACTIVITY_KIND))


def parse_raakvlak_edges(graph: Graph, root: etree.Element):
    elems = root.find("Raakvlakken").iterfind("Raakvlak")
    for el in elems:
        raakvlak = graph[UUID(el.attrib.get("GUID"))]

        rootrefs = [item.attrib["RootRef"] for item in el.iterfind("SI_Objecttype")]

        objecten = []
        for item in el.iterfind("SI_Objecttype"):
            try:
                objecten.append(graph[UUID(item.find("Objecttype").attrib.get("GUID"))])
            except KeyError:
                logger.warning(f"UUID {UUID(item.find('Objecttype').attrib.get('GUID'))} not found")

        functies = [
            graph[UUID(item.find("Functie").attrib.get("GUID"))]
            for item in el.iterfind("SI_Functie")
        ]

        for i, obj in enumerate(objecten):
            annotations = dict(RootRef=rootrefs[i])
            graph.add_edge(Edge(raakvlak, obj, kind=RAAKVLAK_KIND, annotations=annotations))
            graph.add_edge(Edge(obj, raakvlak, kind=RAAKVLAK_KIND, annotations=annotations))
            for func in functies:
                graph.add_edge(Edge(obj, func, kind=RAAKVLAK_KIND, annotations=annotations))
                graph.add_edge(Edge(func, obj, kind=RAAKVLAK_KIND, annotations=annotations))

            for other in objecten[i + 1 :]:
                graph.add_edge(Edge(obj, other, kind=RAAKVLAK_KIND, annotations=annotations))
                graph.add_edge(Edge(other, obj, kind=RAAKVLAK_KIND, annotations=annotations))

        for func in functies:
            graph.add_edge(Edge(raakvlak, func, kind=RAAKVLAK_KIND, annotations=annotations))
            graph.add_edge(Edge(func, raakvlak, kind=RAAKVLAK_KIND, annotations=annotations))


def to_grip(
    graph: Graph, path: Optional[Union[str, Path]] = None, hierarchy_mode: str = "top-down"
) -> Optional[str]:
    """Convert a graph with GRIP content structure to a GRIP XML.

    Arguments:
        graph: Graph to convert.
        path: Optional path to write converted XML text to.

    Returns:
        String contents when no path was given.
    """
    report = _build_report(graph, hierarchy_mode)
    byte = etree.tostring(
        report,
        encoding="UTF-8",
        xml_declaration=True,
        pretty_print=True,
    )
    if path:
        Path(path).write_bytes(byte)
    else:
        return byte.decode()


def _build_report(graph: Graph, hierarchy_mode: str = "top-down") -> etree.Element:
    """Convert Graph object to GRIP Compatible XML.

    Note: Only works for Graphs objects that are created using the from_grip function.

    Arguments:
        graph: Graph object to be converted.
        hierarchy_mode: One of "top-down" or "bottom-up".
            Defines how hierarchy relations between objects are stored within the XML file.

    Returns
        XML report.
    """
    a = graph.annotations
    report = etree.Element("Report")
    report.attrib["ReportName"] = a.get("ReportName", str(uuid4()))
    report.attrib["EnvironmentID"] = a.get("EnvironmentID", str(uuid4()))
    report.attrib["EnvironmentName"] = a.get("EnvironmentName", "Rijkswaterstaat")
    report.attrib["EnvironmentURL"] = a.get(
        "EnvironmentURL", "https://rijkswaterstaat.relaticsonline.com"
    )
    report.attrib["GeneratedOn"] = datetime.now().strftime("%Y-%m-%d")
    report.attrib["WorkspaceID"] = a.get("WorkspaceID", str(uuid4()))
    report.attrib["WorkspaceName"] = a.get("WorkspaceName", report.attrib["WorkspaceID"])
    report.attrib["TargetDevice"] = a.get("TargetDevice", "Pc")

    param_nodes = graph.get_nodes_by_kind(PARAMETER_KIND)
    for cat in REPORT_CATEGORIES:
        el = etree.SubElement(report, cat)
        _add_params(
            el, *[p for p in param_nodes if p.annotations["Name"] in REPORT_CATEGORY_PARAMS[cat]]
        )

    for node in graph.nodes:
        if node.kind == OBJECT_KIND:
            _add_object_node(report.find("Objecten"), node, graph, hierarchy_mode)
        elif node.kind == OBJECTTYPE_KIND:
            _add_objecttype_node(report.find("Objecttypen"), node, graph)
        elif node.kind == FUNCTIE_KIND:
            _add_functie_node(report.find("Functies"), node, graph)
        elif node.kind == SYSTEEMEIS_KIND:
            _add_systeemeis_node(report.find("Systeemeisen"), node, graph)
        elif node.kind == SCOPE_KIND:
            _add_scope_node(report.find("Scope"), node, graph)
        elif node.kind == RAAKVLAK_KIND:
            _add_raakvlak_node(report.find("Raakvlakken"), node, graph)
        elif node.kind == ACTIVITY_KIND:
            _add_activity_node(report.find("Onderhoudsactiviteiten"), node)
        elif node.kind == PARAMETER_KIND:
            pass
        else:
            raise ValueError(f"Don't know this node kind '{node.kind}'")

    return report

def _add_objecttype_node(el: etree.Element, node: Node, graph: Graph):
    """Add objecttype instance to Objecttypen collection in XML.

    Arguments:
      el: Collection to add objecttype node to
      node: The objectttpe node to be added.
      graph: Source graph.
    """
    sub = etree.SubElement(
        el,
        "Objecttype",
        attrib=dict(
            Name=node.annotations["Name"],
            ConfigurationOfRef=node.annotations["ConfigurationOfRef"],
            GUID=node.annotations["GUID"],
            Type="ELEMENT",
        ),
    )

    # Add general information.
    etree.SubElement(sub, "ID", attrib=node.annotations["ID"])
    etree.SubElement(sub, "ExterneCode", attrib=node.annotations["ExterneCode"])

    # Add links to functions
    fcount = 1
    objtcount = 1
    for t in graph.targets_of(node):
        if t.kind == "functie":
            subsub = etree.SubElement(
                sub,
                "SI_Functie",
                attrib=dict(
                    R1Sequence="1",
                    R2Sequence=str(fcount),
                    Type="RELATION_ELEMENT",
                    RootRef=node.annotations["RootRefFunctie"],
                ),
            )
            etree.SubElement(
                subsub,
                t.kind.capitalize(),
                attrib=dict(
                    ConfigurationOfRef=t.annotations["ConfigurationOfRef"],
                    GUID=t.annotations["GUID"],
                ),
            )

            fcount += 1

    # Add inheritance links to other Objecttypes.
    for s in graph.sources_of(node):
        if s.kind == "objecttype" and node.annotations.get("RootRefOnderliggende", None):
            subsub = etree.SubElement(
                sub,
                "SI_Onderliggende",
                attrib=dict(
                    R1Sequence="1",
                    R2Sequence=str(objtcount),
                    Type="RELATION_ELEMENT",
                    RootRef=node.annotations["RootRefOnderliggende"],
                ),
            )

            etree.SubElement(
                subsub,
                "Objecttype",
                attrib=dict(
                    ConfigurationOfRef=s.annotations["ConfigurationOfRef"],
                    GUID=s.annotations["GUID"],
                ),
            )

            objtcount += 1


def _add_object_node(
    el: etree.Element,
    node: Node,
    graph: Graph,
    hierarchy_mode: str = "top-down",
    hierarchy_mode_rootrefs=HIERARCHY_MODE_ROOTREFS,
):
    """Add object instance to Objecten collection in XML.

    Arguments:
      el: Collection to add object node to
      node: The objectttpe node to be added.
      graph: Source graph.
      hierarchy_mode_rootrefs: GRIP specific rootrefs for different hierarchy modes.
    """
    sub = etree.SubElement(
        el,
        "Object",
        attrib=dict(
            Name=node.annotations["Name"],
            ConfigurationOfRef=node.annotations["ConfigurationOfRef"],
            GUID=node.annotations["GUID"],
            Type="ELEMENT",
        ),
    )

    # Add general information.
    etree.SubElement(sub, "ID", attrib=node.annotations["ID"])
    etree.SubElement(sub, "Code", attrib=node.annotations["Code"])

    # Add references to sub-objects in "top-down" mode.
    if hierarchy_mode == "top-down":
        for idx, c in enumerate(node.children):
            subsub = etree.SubElement(
                sub,
                "SI_Onderliggend",
                attrib=dict(
                    R1Sequence="1",
                    R2Sequence=str(idx + 1),
                    Type="RELATION_ELEMENT",
                    RootRef=hierarchy_mode_rootrefs["top-down"],
                ),
            )

            etree.SubElement(
                subsub,
                "ObjectOnderliggend",
                attrib=dict(
                    ConfigurationOfRef=c.annotations["ConfigurationOfRef"],
                    GUID=c.annotations["GUID"],
                ),
            )

    # Add reference to parent in "bottom-up" mode.
    elif hierarchy_mode == "bottom-up" and node.parent:
        subsub = etree.SubElement(
            sub,
            "SI_Bovenliggend",
            attrib=dict(
                Type="RELATION_ELEMENT",
                RootRef=hierarchy_mode_rootrefs["bottom-up"],
            ),
        )

        etree.SubElement(
            subsub,
            "ObjectBovenliggend",
            attrib=dict(
                ConfigurationOfRef=node.parent.annotations["ConfigurationOfRef"],
                GUID=node.parent.annotations["GUID"],
            ),
        )

    # Add relations to functions and object types.
    fcount = 1
    objtcount = 1
    for t in graph.targets_of(node):
        if t.kind == "functie":
            subsub = etree.SubElement(
                sub,
                "SI_Functie",
                attrib=dict(
                    R1Sequence="1",
                    R2Sequence=str(fcount),
                    Type="RELATION_ELEMENT",
                    RootRef="e9dc9736-558d-e211-81b6-001d09fa6b1e",
                ),
            )
            etree.SubElement(
                subsub,
                t.kind.capitalize(),
                attrib=dict(
                    ConfigurationOfRef=t.annotations["ConfigurationOfRef"],
                    GUID=t.annotations["GUID"],
                ),
            )

            fcount += 1

        elif t.kind == "objecttype":
            subsub = etree.SubElement(
                sub,
                "SI_Objecttype",
                attrib=dict(
                    R1Sequence="1",
                    R2Sequence=str(objtcount),
                    Type="RELATION_ELEMENT",
                    RootRef=node.annotations["RootRefObjecttype"],
                ),
            )
            etree.SubElement(
                subsub,
                t.kind.capitalize(),
                attrib=dict(
                    ConfigurationOfRef=t.annotations["ConfigurationOfRef"],
                    GUID=t.annotations["GUID"],
                ),
            )

            objtcount += 1


def add_SI_Functie(sub: etree.Element, func: Dict[str, Any], fcount: int, node: Node):
    """Add function reference to super element.

    Note: This is a bit of a weird quirk of the GRIP data structure.

    Arguments:
        sub: super element to add function element to
        func: function reference.
        fcount: Number of referenced function.
        node: Node to which the reference is added.
    """
    subsub = etree.SubElement(
        sub,
        "SI_Functie",
        attrib=dict(
            R1Sequence="1",
            R2Sequence=str(fcount),
            Type="RELATION_ELEMENT",
            RootRef=node.annotations["RootRefFunctie"],
        ),
    )

    etree.SubElement(
        subsub,
        func.kind.capitalize(),
        attrib=dict(
            ConfigurationOfRef=func.annotations["ConfigurationOfRef"],
            GUID=func.annotations["GUID"],
        ),
    )


def _add_functie_node(el: etree.Element, node: Node, graph: Graph):
    """Add functie instance to Functies collection in XML.

    Arguments:
      el: Collection to add object node to
      node: The objectttpe node to be added.
      graph: Source graph.
    """
    sub = etree.SubElement(
        el,
        "Functie",
        attrib=dict(
            Name=node.annotations["Name"],
            ConfigurationOfRef=node.annotations["ConfigurationOfRef"],
            GUID=node.annotations["GUID"],
            Type="ELEMENT",
        ),
    )

    etree.SubElement(sub, "ID", attrib=node.annotations["ID"])
    etree.SubElement(sub, "ExterneCode", attrib=node.annotations["ExterneCode"])


def _add_CI_MEEisObject(sub: etree.Element, MObj: Dict[str, Any]):
    """Add object reference to super element.

    Note: This is a bit of a weird quirk of the GRIP data structure.

    Arguments:
        sub: super element to add function element to
        Mobj: Object reference.
    """
    subsub = etree.SubElement(
        sub,
        "CI_MEEisObject",
        attrib=dict(
            R1Sequence=MObj["CI_MEEisObject"]["R1Sequence"],
            R2Sequence=MObj["CI_MEEisObject"]["R2Sequence"],
            RootRef=MObj["CI_MEEisObject"]["RootRef"],
            Type=MObj["CI_MEEisObject"]["Type"],
        ),
    )

    subsubsub = etree.SubElement(
        subsub,
        "EisObject",
        attrib=dict(
            Name=MObj["CI_MEEisObject"]["EisObject"]["Name"],
            ConfigurationOfRef=MObj["CI_MEEisObject"]["EisObject"]["ConfigurationOfRef"],
            GUID=MObj["CI_MEEisObject"]["EisObject"]["GUID"],
            Type=MObj["CI_MEEisObject"]["EisObject"]["Type"],
        ),
    )

    subsubsubsub = etree.SubElement(
        subsubsub,
        "SI_Object",
        attrib=dict(
            R1Sequence=MObj["CI_MEEisObject"]["EisObject"]["SI_Object"]["R1Sequence"],
            R2Sequence=MObj["CI_MEEisObject"]["EisObject"]["SI_Object"]["R2Sequence"],
            RootRef=MObj["CI_MEEisObject"]["EisObject"]["SI_Object"]["RootRef"],
            Type=MObj["CI_MEEisObject"]["EisObject"]["SI_Object"]["Type"],
        ),
    )

    etree.SubElement(
        subsubsubsub,
        "Object",
        attrib=MObj["CI_MEEisObject"]["EisObject"]["SI_Object"]["Object"],
    )


def _add_CI_MEEisObjecttype(sub: etree.Element, MObjtype: Dict[str, Any]):
    """Add object reference to super element.

    Note: This is a bit of a weird quirk of the GRIP data structure.

    Arguments:
        sub: super element to add function element to
        Mobjtype: Objecttype reference.
    """
    subsub = etree.SubElement(
        sub,
        "CI_MEEisObjecttype",
        attrib=dict(
            R1Sequence=MObjtype["CI_MEEisObjecttype"]["R1Sequence"],
            R2Sequence=MObjtype["CI_MEEisObjecttype"]["R2Sequence"],
            RootRef=MObjtype["CI_MEEisObjecttype"]["RootRef"],
            Type=MObjtype["CI_MEEisObjecttype"]["Type"],
        ),
    )

    subsubsub = etree.SubElement(
        subsub,
        "EisObjecttype",
        attrib=dict(
            Name=MObjtype["CI_MEEisObjecttype"]["EisObjecttype"]["Name"],
            ConfigurationOfRef=MObjtype["CI_MEEisObjecttype"]["EisObjecttype"][
                "ConfigurationOfRef"
            ],
            GUID=MObjtype["CI_MEEisObjecttype"]["EisObjecttype"]["GUID"],
            Type=MObjtype["CI_MEEisObjecttype"]["EisObjecttype"]["Type"],
        ),
    )

    subsubsubsub = etree.SubElement(
        subsubsub,
        "SI_Objecttype",
        attrib=dict(
            R1Sequence=MObjtype["CI_MEEisObjecttype"]["EisObjecttype"]["SI_Objecttype"][
                "R1Sequence"
            ],
            R2Sequence=MObjtype["CI_MEEisObjecttype"]["EisObjecttype"]["SI_Objecttype"][
                "R2Sequence"
            ],
            RootRef=MObjtype["CI_MEEisObjecttype"]["EisObjecttype"]["SI_Objecttype"]["RootRef"],
            Type=MObjtype["CI_MEEisObjecttype"]["EisObjecttype"]["SI_Objecttype"]["Type"],
        ),
    )

    etree.SubElement(
        subsubsubsub,
        "Objecttype",
        attrib=MObjtype["CI_MEEisObjecttype"]["EisObjecttype"]["SI_Objecttype"]["Objecttype"],
    )


def _add_CI_Eistekst(sub: etree.Element, node: Node):
    """Parse requirement tekst and add them to systeemeis element

    Arguments:
      sub: Systeemeis element
      node: Node to fetch data form.
    """
    if node.annotations.get("CI_EistekstDefinitief", None):
        subsub = etree.SubElement(
            sub,
            "CI_EistekstDefinitief",
            attrib=dict(
                R1Sequence=node.annotations["CI_EistekstDefinitief"]["R1Sequence"],
                R2Sequence=node.annotations["CI_EistekstDefinitief"]["R2Sequence"],
                RootRef=node.annotations["CI_EistekstDefinitief"]["RootRef"],
                Type=node.annotations["CI_EistekstDefinitief"]["Type"],
            ),
        )

        etree.SubElement(subsub, "Eistekst", node.annotations["CI_EistekstDefinitief"]["Eistekst"])

    if node.annotations.get("CI_EistekstOrigineel", None):
        subsub = etree.SubElement(
            sub,
            "CI_EistekstOrigineel",
            attrib=dict(
                R1Sequence=node.annotations["CI_EistekstOrigineel"]["R1Sequence"],
                R2Sequence=node.annotations["CI_EistekstOrigineel"]["R2Sequence"],
                RootRef=node.annotations["CI_EistekstOrigineel"]["RootRef"],
                Type=node.annotations["CI_EistekstOrigineel"]["Type"],
            ),
        )

        etree.SubElement(
            subsub,
            "EistekstOrigineel",
            attrib=node.annotations["CI_EistekstOrigineel"]["EistekstOrigineel"],
        )

def _add_Eistype(sub:etree.Element, node: Node):
    etree.SubElement(
        sub,
        "Eistype",
        attrib=dict(
            Name=node.annotations["SI_Eistype"]["Eistype"]["Name"],
            ConfigurationOfRef=node.annotations["SI_Eistype"]["Eistype"]["ConfigurationOfRef"],
            GUID=node.annotations["SI_Eistype"]["Eistype"]["GUID"]
        )
    )


def _add_SI_Eistype(sub:etree.Element, node: Node):
    subsub = etree.SubElement(
        sub,
        "SI_Eistype",
        attrib=dict(
            R1Sequence=node.annotations["SI_Eistype"]["R1Sequence"],
            R2Sequence=node.annotations["SI_Eistype"]["R2Sequence"],
            RootRef=node.annotations["SI_Eistype"]["RootRef"],
            Type=node.annotations["SI_Eistype"]["Type"],
        )
    )

    _add_Eistype(sub=subsub, node=node)

def _add_Periode(sub:etree.Element, node: Node):
    etree.SubElement(
        sub,
        "Periode",
        attrib=dict(
            Name=node.annotations["SI_Periode"]["Periode"]["Name"],
            ConfigurationOfRef=node.annotations["SI_Periode"]["Periode"]["ConfigurationOfRef"],
            GUID=node.annotations["SI_Periode"]["Periode"]["GUID"]
        )
    )


def _add_SI_Periode(sub:etree.Element, node: Node):
    subsub = etree.SubElement(
        sub,
        "SI_Periode",
        attrib=dict(
            R1Sequence=node.annotations["SI_Periode"]["R1Sequence"],
            R2Sequence=node.annotations["SI_Periode"]["R2Sequence"],
            RootRef=node.annotations["SI_Periode"]["RootRef"],
            Type=node.annotations["SI_Periode"]["Type"],
        )
    )

    _add_Periode(sub=subsub, node=node)


def _add_Systeemeisgroepering(sub:etree.Element, group: Dict):
    etree.SubElement(
        sub,
        "Systeemeisgroepering",
        attrib=dict(
            Name=group["Name"],
            ConfigurationOfRef=group["ConfigurationOfRef"],
            GUID=group["GUID"]
        )
    )


def _add_SI_Systeemeisgroepering(sub:etree.Element, sg: Dict):
    subsub = etree.SubElement(
        sub,
        "SI_Systeemeisgroepering",
        attrib=dict(
            R1Sequence=sg["R1Sequence"],
            R2Sequence=sg["R2Sequence"],
            RootRef=sg["RootRef"],
            Type=sg["Type"],
        )
    )

    _add_Systeemeisgroepering(sub=subsub, group=sg["Systeemeisgroepering"])


def _add_Methode(sub:etree.Element, v: Dict):
    etree.SubElement(
        sub,
        "Methode",
        attrib=dict(
            Name=v["SI_Methode"]["Methode"]["Name"],
            ConfigurationOfRef=v["SI_Methode"]["Methode"]["ConfigurationOfRef"],
            GUID=v["SI_Methode"]["Methode"]["GUID"]
        )
    )


def _add_SI_Methode(sub:etree.Element, v: Dict):
    subsub = etree.SubElement(
        sub,
        "SI_Methode",
        attrib=dict(
            R1Sequence=v["SI_Methode"]["R1Sequence"],
            R2Sequence=v["SI_Methode"]["R2Sequence"],
            RootRef=v["SI_Methode"]["RootRef"],
            Type=v["SI_Methode"]["Type"],
        )
    )

    _add_Methode(subsub, v)
    
def _add_Fase(sub:etree.Element, v: Dict):
    etree.SubElement(
        sub,
        "Fase",
        attrib=dict(
            Name=v["SI_Fase"]["Fase"]["Name"],
            ConfigurationOfRef=v["SI_Fase"]["Fase"]["ConfigurationOfRef"],
            GUID=v["SI_Fase"]["Fase"]["GUID"]
        )
    )


def _add_SI_Fase(sub:etree.Element, v: Dict):
    subsub = etree.SubElement(
        sub,
        "SI_Fase",
        attrib=dict(
            R1Sequence=v["SI_Fase"]["R1Sequence"],
            R2Sequence=v["SI_Fase"]["R2Sequence"],
            RootRef=v["SI_Fase"]["RootRef"],
            Type=v["SI_Fase"]["Type"],
        )
    )

    _add_Fase(sub=subsub, v=v)


def _add_Toelichting(sub:etree.Element, v: Dict):
    etree.SubElement(
        sub,
        "Toelichting",
        attrib=dict(
            Name=v["CI_Toelichting"]["Toelichting"]["Name"],
            ConfigurationOfRef=v["CI_Toelichting"]["Toelichting"]["ConfigurationOfRef"],
            GUID=v["CI_Toelichting"]["Toelichting"]["GUID"],
            Description=v["CI_Toelichting"]["Toelichting"]["Description"],
        )
    )


def _add_CI_Toelichting(sub:etree.Element, v: Dict):
    subsub = etree.SubElement(
        sub,
        "CI_Toelichting",
        attrib=dict(
            R1Sequence=v["CI_Toelichting"]["R1Sequence"],
            R2Sequence=v["CI_Toelichting"]["R2Sequence"],
            RootRef=v["CI_Toelichting"]["RootRef"],
            Type=v["CI_Toelichting"]["Type"],
        )
    )

    _add_Toelichting(sub=subsub, v=v)

def _add_Criterium(sub:etree.Element, v: Dict):
    etree.SubElement(
        sub,
        "Criterium",
        attrib=v["Criterium"]
    )


def _add_Verificatie(sub:etree.Element, v: Dict):
    subsub = etree.SubElement(
        sub,
        "Verificatie",
        attrib=dict(
            Name=v["Name"],
            ConfigurationOfRef=v["ConfigurationOfRef"],
            GUID=v["GUID"],
        )
    )

    _add_Criterium(sub=subsub, v=v)
    _add_SI_Methode(sub=subsub, v=v)
    _add_SI_Fase(sub=subsub, v=v)
    _add_CI_Toelichting(sub=subsub, v=v)
    


def _add_CI_Verificatievoorschrift(sub:etree.Element, vv: Dict):
    subsub = etree.SubElement(
        sub,
        "CI_Verificatievoorschrift",
        attrib=dict(
            R1Sequence=vv["R1Sequence"],
            R2Sequence=vv["R2Sequence"],
            RootRef=vv["RootRef"],
            Type=vv["Type"],
        )
    )

    _add_Verificatie(sub=subsub, v=vv["Verificatie"])


def _add_systeemeis_node(el: etree.Element, node: Node, graph: Graph):
    """Add systemeisen instance to Systeemeisen collection in XML.

    Arguments:
      el: Collection to add object node to
      node: Node to fetch data from.
    """
    sub = etree.SubElement(
        el,
        "Systeemeis",
        attrib=dict(
            Name=node.annotations["Name"],
            ConfigurationOfRef=node.annotations["ConfigurationOfRef"],
            GUID=node.annotations["GUID"],
            Type="ELEMENT",
        ),
    )

    etree.SubElement(sub, "ID", attrib=node.annotations["ID"])
    etree.SubElement(sub, "BronID", attrib=node.annotations["BronID"])
    etree.SubElement(sub, "Eiscodering", attrib=node.annotations["Eiscodering"])

    _add_CI_Eistekst(sub=sub, node=node)

    # Add references to objects
    for MObj in node.annotations.get("CI_MEEisObjects", []):
        _add_CI_MEEisObject(sub=sub, MObj=MObj)

    # Add references to objectypes.
    for MObjtype in node.annotations.get("CI_MEEisObjecttypen", []):
        _add_CI_MEEisObjecttype(sub=sub, MObjtype=MObjtype)

    # Add requirement type.
    if node.annotations.get("SI_Eistype", None):
        _add_SI_Eistype(sub=sub, node=node)    

     # Add requirement period.
    if node.annotations.get("SI_Periode", None):
        _add_SI_Periode(sub=sub, node=node)  

    # Add requirement period.
    if node.annotations.get("SI_Periode", None):
        _add_SI_Periode(sub=sub, node=node)  

    # Add requirement groepering.
    for sg in node.annotations.get("SI_Systeemeisgroepering", []):
        _add_SI_Systeemeisgroepering(sub=sub, sg=sg)  

    for vv in node.annotations.get("CI_Verificatievoorschrift", []):
        _add_CI_Verificatievoorschrift(sub=sub, vv=vv)

    if node.annotations.get("CI_Toelichting"):
        _add_CI_Toelichting(sub=sub, v=node.annotations)

    # Add references to functions.
    fcount = 1
    for t in graph.targets_of(node):
        if t.kind == "functie":
            add_SI_Functie(sub=sub, func=t, fcount=fcount, node=node)
            fcount += 1


def _add_scope_node(el: etree.Element, node: Node, graph: Graph):
    """Add scope instance to Scope collection in XML.

    Arguments:
      el: Collection to add object node to
      node: Node to fetch data from.
    """
    sub = etree.SubElement(
        el,
        "Scope",
        attrib=dict(
            Name=node.annotations["Name"],
            GUID=node.annotations["GUID"],
            ConfigurationOfRef=node.annotations["ConfigurationOfRef"],
            Type=node.annotations["Type"],
        ),
    )

    counts = dict(functie=0, raakvlak=0, systeemeis=0, object=0)

    # Add relations to elements part of scope.
    for e in graph.edges_from(node):
        t = e.target

        counts[t.kind] += 1

        subsub = etree.SubElement(
            sub,
            f"SI_{t.kind.capitalize()}",
            attrib=dict(
                R1Sequence="1",
                R2Sequence=str(counts[t.kind]),
                Type="RELATION_ELEMENT",
                RootRef=e.annotations["RootRef"],
            ),
        )
        etree.SubElement(
            subsub,
            t.kind.capitalize(),
            attrib={
                t.kind.capitalize(): t.annotations["Name"],
                "ConfigurationOfRef": t.annotations["ConfigurationOfRef"],
                "GUID": t.annotations["GUID"],
            },
        )


def _add_raakvlak_node(el: etree.Element, node: Node, graph: Graph):
    """Add raakvlak instance to Raakvlakken collection in XML.

    Arguments:
      el: Collection to add object node to
      node: Node to fetch data from.
      graph: Graph to fetch data from.
    """
    sub = etree.SubElement(
        el,
        "Raakvlak",
        attrib=dict(
            Name=node.annotations["Name"],
            Description=node.annotations["Description"],
            ConfigurationOfRef=node.annotations["ConfigurationOfRef"],
            GUID=node.annotations["GUID"],
            Type="ELEMENT",
        ),
    )

    etree.SubElement(sub, "ID", attrib=node.annotations["ID"])

    etree.SubElement(sub, "BronID", attrib=node.annotations["BronID"])

    count = 1
    for t in graph.targets_of(node):
        if t.kind != "object":
            continue

        subsub = etree.SubElement(
            sub,
            f"SI_{t.kind.capitalize()}type",
            attrib=dict(
                R1Sequence="1",
                R2Sequence=str(count),
                Type="RELATION_ELEMENT",
                RootRef=[e for e in graph[node.name, t.name]][0].annotations["RootRef"],
            ),
        )
        etree.SubElement(
            subsub,
            f"{t.kind.capitalize()}type",
            attrib=dict(
                ConfigurationOfRef=t.annotations["ConfigurationOfRef"],
                GUID=t.annotations["GUID"],
            ),
        )

        count += 1


def _add_activity_node(el: etree.Element, node: Node):
    """Add activity instance to Onderhoudsactiviteiten collection in XML.

    Arguments:
      el: Collection to add object node to
      node: Node to fetch data from.
      graph: Graph to fetch data from.
    """
    sub = etree.SubElement(
        el,
        "Onderhoudsactiviteit",
        attrib=dict(
            Name=node.annotations["Name"],
            ConfigurationOfRef=node.annotations["ConfigurationOfRef"],
            GUID=node.annotations["GUID"],
            Type="ELEMENT",
        ),
    )

    etree.SubElement(
        sub, 
        "ID", 
        attrib=dict(
            IDOA=node.annotations["ID"]["IDOA"],
            RootRef=node.annotations["ID"]["RootRef"]
        )
    )

    etree.SubElement(
        sub,
        "Onderhoudsactiviteittype", 
        attrib=node.annotations["Onderhoudsactiviteittype"]
    )

    etree.SubElement(
        sub,
        "Aard", 
        attrib=node.annotations["Aard"]
    )

    for MOa in node.annotations["CI_MEOnderhoudsactiviteitObjecten"]:
        _add_CI_MEOnderhoudsactiviteitObject(sub, MOa)


def _add_CI_MEOnderhoudsactiviteitObject(sub: etree.Element, MObj: Dict[str, Any]):
    """Add object reference to Object.

    Note: This is a bit of a weird quirk of the GRIP data structure.

    Arguments:
        sub: super element to add function element to
        Mobj: Object reference.
    """
    subsub = etree.SubElement(
        sub,
        "CI_MEOnderhoudsactiviteitObject",
        attrib=dict(
            R1Sequence=MObj["CI_MEOnderhoudsactiviteitObject"]["R1Sequence"],
            R2Sequence=MObj["CI_MEOnderhoudsactiviteitObject"]["R2Sequence"],
            RootRef=MObj["CI_MEOnderhoudsactiviteitObject"]["RootRef"],
            Type=MObj["CI_MEOnderhoudsactiviteitObject"]["Type"],
        ),
    )

    subsubsub = etree.SubElement(
        subsub,
        "MEOnderhoudsactiviteitObject",
        attrib=dict(
            Name=MObj["CI_MEOnderhoudsactiviteitObject"]["MEOnderhoudsactiviteitObject"]["Name"],
            ConfigurationOfRef=MObj["CI_MEOnderhoudsactiviteitObject"]["MEOnderhoudsactiviteitObject"]["ConfigurationOfRef"],
            GUID=MObj["CI_MEOnderhoudsactiviteitObject"]["MEOnderhoudsactiviteitObject"]["GUID"],
            Type=MObj["CI_MEOnderhoudsactiviteitObject"]["MEOnderhoudsactiviteitObject"]["Type"],
        ),
    )

    subsubsubsub = etree.SubElement(
        subsubsub,
        "SI_Object",
        attrib=dict(
            R1Sequence=MObj["CI_MEOnderhoudsactiviteitObject"]["MEOnderhoudsactiviteitObject"]["SI_Object"]["R1Sequence"],
            R2Sequence=MObj["CI_MEOnderhoudsactiviteitObject"]["MEOnderhoudsactiviteitObject"]["SI_Object"]["R2Sequence"],
            RootRef=MObj["CI_MEOnderhoudsactiviteitObject"]["MEOnderhoudsactiviteitObject"]["SI_Object"]["RootRef"],
            Type=MObj["CI_MEOnderhoudsactiviteitObject"]["MEOnderhoudsactiviteitObject"]["SI_Object"]["Type"],
        ),
    )

    etree.SubElement(
        subsubsubsub,
        "Object",
        attrib=MObj["CI_MEOnderhoudsactiviteitObject"]["MEOnderhoudsactiviteitObject"]["SI_Object"]["Object"],
    )


def _add_params(el: etree.Element, *params: Node):
    """Add parameter instance to Paramer collection in XML.

    Arguments:
      el: Collection to add object node to
      params: Nodes to fetch data from.
    """
    sub = etree.SubElement(el, "RelaticsParameters")
    for p in params:
        etree.SubElement(sub, "RelaticsParameter", attrib=p.annotations)


def add_raakvlak_annotations(g: Graph):
    """Adding Raakvlak ID's as a string to edges annotations."""
    for e in g.edges:
        if e.source.kind != "object" or e.target.kind != "object":
            continue

        sr = set([t for t in g.targets_of(e.source) if t.kind == "raakvlak"])
        tr = set([t for t in g.targets_of(e.target) if t.kind == "raakvlak"])

        if not sr.intersection(tr):
            e.annotations.raakvlakken = ""
            continue

        raakvlakken = ",".join([r.name.split(" | ")[-1] for r in sr.intersection(tr)])

        e.annotations.raakvlakken = raakvlakken


def add_labels(g: Graph):
    """Add labels to edges based on the description of Raakvlakken."""
    for n in [node for node in g.nodes if node.kind == "raakvlak"]:
        if not n.annotations.get("Description", None):
            # No labels to derive.
            continue
        labels = [label.strip() for label in n.annotations.Description.split(",")]
        variants = [s for s in g.sources_of(n)] + [t for t in g.targets_of(n)]
        for e in g.edges_between_all(variants, variants):
            e.labels = list(set(e.labels + labels))

            if len(e.labels) >= 1 and "default" in e.labels:
                e.labels.remove("default")



from ragraph import datasets
from ragraph.analysis import sequence


def test_axis_sequencing_climate_control_mg():
    g = datasets.get("climate_control_mg")

    g, seq = sequence.axis(g, nodes=g.leafs, names=True)

    assert seq == [
        "Compressor",
        "Air Controls",
        "Command Distribution",
        "Evaporator Case",
        "Evaporator Core",
        "Blower Motor",
        "Engine Fan",
        "Radiator",
        "Condenser",
        "Heater Core",
        "Heater Hoses",
        "Accumulator",
        "Refrigeration Controls",
        "Actuators",
        "Blower Controller",
        "Sensors",
    ]



from ragraph.analysis._utils import create_parent
from ragraph.analysis.cluster import tarjans_scc
from ragraph.analysis.sequence import branchsort, scc_tearing
from ragraph.io.matrix import from_matrix


def test_sequence():
    g = from_matrix(
        [
            [0, 1, 1, 0, 0],  # 0 -> 4, only 1 & 2 as input.
            [0, 0, 0, 1, 0],  # 1 -> 2, only 3 as input.
            [0, 1, 0, 0, 0],  # 2 -> 3, only 1 as input.
            [0, 0, 0, 0, 0],  # 3 -> 1, no inputs
            [0, 0, 0, 0, 0],  # 4 -> 0, no inputs
        ]
    )

    # Load the graph.
    _, seq = tarjans_scc(g, names=True)
    assert seq[::-1] == [
        "node4",
        "node3",
        "node1",
        "node2",
        "node0",
    ], "Trivial sequence should be found by SCC reversed."

    create_parent(g, children=["node2", "node0"])
    _, seq = tarjans_scc(g, leafs=g.roots, names=True)
    assert seq[::-1] == [
        "node4",
        "node3",
        "node1",
        "node.node0",
    ], "SCC should work on nested graphs."

    g, root_seq, leaf_seq = branchsort(g, nodes=g.roots, algo=scc_tearing, names=True)
    assert root_seq == [
        "node4",
        "node3",
        "node1",
        "node.node0",
    ], "Branchsort should return trivial sub-sequences correctly."
    assert leaf_seq == [
        "node4",
        "node3",
        "node1",
        "node2",
        "node0",
    ], "Branchsort should return trivial sub-sequences correctly."



import difflib
import shutil
import uuid
from pathlib import Path

import pytest

from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node

HERE = Path(__file__).parent
UPDATE_PATH = HERE / ".update"
UPDATE = UPDATE_PATH.exists()
INSPECT_PATH = HERE / ".inspect"
INSPECT = INSPECT_PATH.exists()
SPECS_PATH = HERE / "data" / "specs"
DOCS_DIR = HERE.parent / "docs"


@pytest.fixture(autouse=True, scope="session")
def generated_docs():
    path = DOCS_DIR / "generated"
    shutil.rmtree(path, ignore_errors=True)
    path.mkdir(exist_ok=True)


# Fix UUID generation during tests:
def generate_int_uuids():
    index = 0
    while True:
        yield uuid.UUID(int=index)
        index += 1


@pytest.fixture
def inspect():
    return INSPECT


@pytest.fixture(autouse=True, scope="function")
def reset_uuids():
    int_uuid = generate_int_uuids()
    uuid.uuid4 = lambda: next(int_uuid)


@pytest.fixture
def update():
    return UPDATE


@pytest.fixture
def datadir():
    return Path(__file__).parent / "data"


@pytest.fixture
def a():
    return Node("a")


@pytest.fixture
def b():
    return Node("b")


@pytest.fixture
def c():
    return Node("c")


@pytest.fixture
def d():
    return Node("d")


@pytest.fixture
def e():
    return Node("e")


@pytest.fixture
def f():
    return Node("f")


@pytest.fixture
def empty_graph():
    return Graph()


@pytest.fixture
def edge_ab(a, b):
    return Edge(a, b, name="a->b")


@pytest.fixture
def graph_ab(a, b, edge_ab):
    return Graph(nodes=[a, b], edges=[edge_ab], name="graph_ab")


@pytest.fixture
def rich_graph(graph_ab):
    graph = graph_ab
    graph.name = "rich"
    c = Node("c", parent=graph.node_dict["a"], kind="rich", weights=dict(dollars=1))
    d = Node("d", parent=graph.node_dict["a"], kind="wealthy", weights=dict(euros=0.5))
    graph.add_node(c)
    graph.add_node(d)
    graph.add_edge(Edge(c, d, name="c->d", kind="money", labels=["euros"], weights=dict(dollars=1)))
    return graph


@pytest.fixture
def esl_file(datadir):
    return datadir / "esl/test.esl"


@pytest.fixture
def esl_graph(esl_file):
    from ragraph.io.esl import from_esl

    return from_esl(esl_file)


@pytest.fixture
def check_diff():
    """Check whether there is a diff w.r.t. to the reference path."""

    def _check_diff(path: Path, ref_path: Path, sort: bool = False):
        text = path.read_text(encoding="utf-8") if path.exists() else ""
        ref_text = ref_path.read_text(encoding="utf-8") if path.exists() else ""

        lines = text.splitlines(False)
        ref_lines = ref_text.splitlines(False)
        if sort:
            lines, ref_lines = sorted(lines), sorted(ref_lines)

        diff = difflib.unified_diff(
            ref_lines,
            lines,
            fromfile=str(ref_path),
            tofile=str(path),
            lineterm="",
        )
        diffstr = "\n".join(diff)

        if diffstr and UPDATE:
            if not ref_path.parent.exists():
                ref_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(path, ref_path)
        else:
            assert not diffstr, diffstr

    return _check_diff


@pytest.fixture
def check_plotly(datadir: Path, inspect: bool):
    """Plotly output check function."""
    from plotly import io as pio
    from plotly.graph_objs import Figure

    def _check_plotly(fig: Figure, fname: str):
        """Checking if the figure data and shapes are equal to the data and shapes
        stored in the reference file.

        Arguments:
            fig: The figure to be tested.
            fname: The relative path to the reference file.
        """
        if inspect:
            fig.show()

        fpath = datadir / "plotly" / fname
        if UPDATE:
            pio.write_json(fig, fpath)
        reference = pio.read_json(fpath)

        assert fig == reference, "Figures should match."

    return _check_plotly



from ragraph import datasets
from ragraph.analysis.sequence._genetic import genetic, genetic_sequencing


def test_sga_seq():
    graph = datasets.get("ucav")
    nodes = list(graph.nodes)

    lin = genetic_sequencing(graph, nodes, 100, 100, 5, 0.3, 0.05, 0.1)

    best = lin.hall_of_fame.chromosomes[0].genes
    _ = [nodes[i] for i in best]

    graph, seq2 = genetic(
        graph,
        nodes=nodes,
        n_chromosomes=100,
        n_generations=100,
        n_hall_of_fame=5,
        p_crossover=0.3,
        p_mutation=0.05,
        p_swap=0.1,
    )



"""Tests for the plot components module."""

import json

import pytest
from plotly import graph_objs as go

from ragraph import datasets
from ragraph.datasets import get
from ragraph.plot import components
from ragraph.plot.generic import Style


@pytest.fixture
def leafs():
    g = datasets.get("climate_control")

    leafs = sorted(g.leafs, key=lambda x: x.name)
    return leafs


@pytest.fixture
def edges():
    g = datasets.get("climate_control")
    edges = g.edges
    return edges


def test_labels(leafs, tmpdir, datadir, check_diff):
    labels = components.Labels(leafs=leafs)
    fname = "labels_test_expected.json"

    d = labels.as_dict()
    data = dict(width=d["width"], height=d["height"], annotations=d["annotations"])
    (tmpdir / fname).write_text(
        json.dumps(data, sort_keys=True, indent=2),
        encoding="utf-8",
    )

    check_diff(tmpdir / fname, datadir / "json" / fname, sort=True)


def test_empty_labels():
    components.Labels(leafs=[])


def test_tree(leafs):
    tree = components.Tree(leafs=leafs)
    assert tree.as_dict() == {
        "width": 20,
        "height": 320,
        "annotations": [],
        "traces": [
            go.Scatter(
                {
                    "customdata": [
                        {"data": "Accumulator", "pointtype": "tree_node"},
                        {"data": "Actuators", "pointtype": "tree_node"},
                        {"data": "Air Controls", "pointtype": "tree_node"},
                        {"data": "Blower Controller", "pointtype": "tree_node"},
                        {"data": "Blower Motor", "pointtype": "tree_node"},
                        {"data": "Command Distribution", "pointtype": "tree_node"},
                        {"data": "Compressor", "pointtype": "tree_node"},
                        {"data": "Condenser", "pointtype": "tree_node"},
                        {"data": "Engine Fan", "pointtype": "tree_node"},
                        {"data": "Evaporator Case", "pointtype": "tree_node"},
                        {"data": "Evaporator Core", "pointtype": "tree_node"},
                        {"data": "Heater Core", "pointtype": "tree_node"},
                        {"data": "Heater Hoses", "pointtype": "tree_node"},
                        {"data": "Radiator", "pointtype": "tree_node"},
                        {"data": "Refrigeration Controls", "pointtype": "tree_node"},
                        {"data": "Sensors", "pointtype": "tree_node"},
                    ],
                    "hoverinfo": "text",
                    "marker": {"color": "gray"},
                    "mode": "markers",
                    "showlegend": False,
                    "text": [
                        "Accumulator",
                        "Actuators",
                        "Air Controls",
                        "Blower Controller",
                        "Blower Motor",
                        "Command Distribution",
                        "Compressor",
                        "Condenser",
                        "Engine Fan",
                        "Evaporator Case",
                        "Evaporator Core",
                        "Heater Core",
                        "Heater Hoses",
                        "Radiator",
                        "Refrigeration Controls",
                        "Sensors",
                    ],
                    "x": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    "y": [
                        15.5,
                        14.5,
                        13.5,
                        12.5,
                        11.5,
                        10.5,
                        9.5,
                        8.5,
                        7.5,
                        6.5,
                        5.5,
                        4.5,
                        3.5,
                        2.5,
                        1.5,
                        0.5,
                    ],
                }
            )
        ],
        "shapes": [],
        "xaxis": go.layout.XAxis(
            {
                "automargin": False,
                "autorange": False,
                "range": [-0.5, 0.5],
                "scaleanchor": "y",
                "scaleratio": 1.0,
                "showgrid": False,
                "showline": False,
                "showticklabels": False,
                "zeroline": False,
            }
        ),
        "yaxis": go.layout.YAxis(
            {
                "automargin": False,
                "autorange": False,
                "range": [0, 16.0],
                "scaleanchor": "x",
                "scaleratio": 1.0,
                "showgrid": True,
                "showline": False,
                "showticklabels": False,
                "zeroline": False,
            }
        ),
    }


def test_legend(update):
    g = get("elevator45")

    # Categorical legend.
    style = Style(
        piemap=dict(
            display="labels",
            mode="relative",
        ),
    )

    legend = components.Legend(edges=g.edges, style=style)

    assert legend.as_dict() == {
        "width": 34.412,
        "height": 100,
        "annotations": [],
        "traces": [
            go.Scatter(
                {
                    "hoverinfo": "text",
                    "marker": {"color": "#6a9ffc"},
                    "mode": "markers",
                    "showlegend": False,
                    "text": ["EE"],
                    "x": [0.5],
                    "y": [4.5],
                }
            ),
            go.Scatter(
                {
                    "hoverinfo": "text",
                    "marker": {"color": "#f27663"},
                    "mode": "markers",
                    "showlegend": False,
                    "text": ["ME"],
                    "x": [0.5],
                    "y": [3.5],
                }
            ),
            go.Scatter(
                {
                    "hoverinfo": "text",
                    "marker": {"color": "#56c950"},
                    "mode": "markers",
                    "showlegend": False,
                    "text": ["i"],
                    "x": [0.5],
                    "y": [2.5],
                }
            ),
            go.Scatter(
                {
                    "hoverinfo": "text",
                    "marker": {"color": "#a57fff"},
                    "mode": "markers",
                    "showlegend": False,
                    "text": ["m"],
                    "x": [0.5],
                    "y": [1.5],
                }
            ),
            go.Scatter(
                {
                    "hoverinfo": "text",
                    "marker": {"color": "#c89336"},
                    "mode": "markers",
                    "showlegend": False,
                    "text": ["s"],
                    "x": [0.5],
                    "y": [0.5],
                }
            ),
            go.Scatter(
                {
                    "hoverinfo": "text",
                    "mode": "text",
                    "showlegend": False,
                    "text": ["EE", "ME", "i", "m", "s"],
                    "textfont": {
                        "color": "black",
                        "family": "Fira Code,Hack,Courier New,monospace",
                        "size": 12,
                    },
                    "textposition": "middle right",
                    "x": [1.0, 1.0, 1.0, 1.0, 1.0],
                    "y": [4.5, 3.5, 2.5, 1.5, 0.5],
                }
            ),
        ],
        "shapes": [],
        "xaxis": go.layout.XAxis(
            {
                "automargin": False,
                "range": [0, 1.7206],
                "scaleanchor": "y",
                "showgrid": False,
                "showline": False,
                "showticklabels": False,
                "zeroline": False,
            }
        ),
        "yaxis": go.layout.YAxis(
            {
                "automargin": False,
                "autorange": False,
                "range": [0, 6],
                "scaleratio": 1.0,
                "showgrid": False,
                "showline": False,
                "showticklabels": False,
                "zeroline": False,
            }
        ),
    }

    # Numerical legend.
    style = Style(
        piemap=dict(
            display="weights",
            mode="relative",
            fields=["sum"],
        ),
        legend=dict(height=10),
    )

    legend = components.Legend(edges=g.edges, style=style)

    assert legend.width == 50
    assert legend.height == 200


def test_piemap(leafs, edges):
    piemap = components.PieMap(rows=leafs, cols=leafs, edges=edges)
    assert len(piemap.traces) == len(edges)
    assert len(piemap.shapes) >= len(edges)



"""Tests for generic Metadata."""

from ragraph import datasets
from ragraph.generic import Bound, MetadataFilter


def test_metadata_filter(rich_graph):
    empty = MetadataFilter()
    assert not empty.get_checks(), "No settings means no checks."
    assert empty.filter(rich_graph.node_list) == rich_graph.node_list

    node_kinds = MetadataFilter(kinds={"node", "rich"})
    assert len(node_kinds.get_checks()) == 1
    assert [n.name for n in node_kinds.filter(rich_graph.node_list)] == ["a", "b", "c"]

    edge_labels_annots = MetadataFilter(labels={"money", "us"}, annotations={"comment"})
    assert len(edge_labels_annots.get_checks()) == 2
    assert [e.name for e in edge_labels_annots.filter(rich_graph.edge_list)] == []


def test_metadata_weights_filter():
    g = datasets.get("climate_control")
    mf = MetadataFilter(
        weight_domains=dict(spatial=(Bound(-2, inclusive=True), Bound(0, inclusive=False)))
    )
    assert mf.filter(g.edges) == []



import pytest
from ragraph.analysis import sequence
from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node


def test_dfs():
    a = Node("a")
    b = Node("b")
    c = Node("c")
    g = Graph(nodes=[a, b, c], edges=[Edge(a, c), Edge(a, b)])

    _, seq = sequence.tarjans_dfs(g, names=True)
    assert seq == ["a", "b", "c"]

    with pytest.raises(ValueError):
        g = Graph(nodes=[a, b], edges=[Edge(a, b), Edge(b, a)])  # Not a DAG.
        sequence.tarjans_dfs(g)


def test_dfs_chain(chain_graph):
    _, seq = sequence.tarjans_dfs(chain_graph, names=True)
    assert seq == ["4", "3", "2", "1", "0"]


def test_dfs_cases(case: Graph):
    try:
        sequence.tarjans_dfs(case)
    except ValueError as e:
        if str(e) == "Not a Directed Acyclic Graph (DAG).":
            pass
        else:
            raise e



from ragraph.io.yaml import from_yaml, to_yaml


def test_yaml(rich_graph, datadir, tmpdir, update):
    tmp = tmpdir / "test.yaml"
    ref = datadir / "yaml/test.yaml"

    if update:
        to_yaml(rich_graph, ref)

    # Test loading
    ref_data = from_yaml(ref)
    assert ref_data.json_dict == rich_graph.json_dict

    # Test a round trip
    to_yaml(rich_graph, tmp)
    round_trip = from_yaml(tmp)
    assert round_trip.json_dict == rich_graph.json_dict



"""Tests for generic utils."""

import pytest
from ragraph import utils
from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node


@pytest.fixture
def depgraph():
    a = Node("a", kind="lead")
    b = Node("b", kind="lead")
    c = Node("c", kind="other")
    d = Node("d", kind="other")
    e = Node("e", kind="other")
    f = Node("f", kind="other")

    ac = Edge(a, c, kind="include")
    db = Edge(d, b, kind="include")
    ae = Edge(a, e, kind="exclude")
    fb = Edge(f, b, kind="exclude")

    g = Graph(nodes=[a, b, c, d, e, f], edges=[ac, db, ae, fb])
    return g


def test_is_dependent(depgraph: Graph):
    lead = [n for n in depgraph.nodes if n.kind == "lead"]
    other = [n for n in depgraph.nodes if n.kind == "other"]

    assert [utils.is_dependent(depgraph, lead, n, {"include"}) for n in other] == [
        True,
        True,
        False,
        False,
    ]


def test_get_up_to_depth(rich_graph: Graph):
    res = list(utils.get_up_to_depth([rich_graph["a"]], 0))
    assert res == [rich_graph["a"]]
    res = list(utils.get_up_to_depth([rich_graph["a"]], 1))
    assert res == [rich_graph[i] for i in ["c", "d"]]


def test_select_nodes(esl_graph: Graph):
    nodes = utils.select_nodes(
        esl_graph,
        ["component", "variable"],
        [],
        depth=1,
        selection_mode="dependent",
    )
    assert [n.name for n in nodes] == ["world.t", "world.q", "world.r"]

    nodes = utils.select_nodes(
        esl_graph,
        ["component", "variable_type"],
        [],
        depth=2,
        selection_mode="dependent",
    )
    assert [n.name for n in nodes] == ["world.t", "world.q"]

    nodes = utils.select_nodes(
        esl_graph,
        ["component", "variable_type"],
        [],
        depth=2,
        selection_mode="independent",
    )
    assert [n.name for n in nodes] == [
        "world.t",
        "world.q",
        "real",
        "integer",
        "string",
        "boolean",
    ]


def test_select_nodes_bad_mode(depgraph: Graph):
    with pytest.raises(ValueError) as e:
        utils.select_nodes(depgraph, [], [], depth=2, selection_mode="bad mode")
    assert "selection mode" in str(e.value)


def test_graph_get_node_selection(esl_graph):
    ures = utils.select_nodes(
        esl_graph,
        ["component", "variable_type"],
        [],
        depth=2,
        selection_mode="independent",
    )
    gres = esl_graph.get_node_selection(
        ["component", "variable_type"], [], depth=2, selection_mode="independent"
    )
    assert ures == gres



from ragraph import colors
from ragraph.plot import utils
from ragraph.plot.generic import Component


def test_get_subplots():
    cmps = [
        [Component(width=100, height=100), Component(width=300, height=200)],
        [Component(width=50, height=50), Component(height=200)],
    ]
    fig = utils.get_subplots(cmps)
    assert fig
    assert fig.layout.xaxis.domain == (0.0, 0.5)
    assert fig.layout.yaxis2.domain == (0.5, 1.0)


def test_swatchplot():
    from plotly import graph_objs as go

    fig = utils.get_swatchplot(cat=colors.get_categorical(), con=colors.get_continuous())

    assert isinstance(fig.data[0], go.Bar)
    assert isinstance(fig.data[1], go.Bar)



from pathlib import Path

import pytest
from ragraph.io.json import from_json, to_json


def test_simple_json(datadir, tmpdir, update):
    path = datadir / "json" / "simple.json"
    g = from_json(path=path, use_uuid=True)
    if update:
        path.write_text(to_json(g))
    assert g.node_count == 6
    assert g.edge_count == 12
    assert to_json(g) == path.read_text()
    to_json(g, (Path(tmpdir) / "simple.json"))  # Test writing to file.


def test_rich_json(datadir, update):
    path = datadir / "json" / "rich.json"
    g = from_json(path=path, use_uuid=True)
    if update:
        path.write_text(to_json(g))
    assert g.node_kinds == ["broke", "dagobert", "empty", "poor", "rich", "wealthy"]
    assert g.node_weight_labels == ["bank"]
    assert g.edge_kinds == ["dollars", "euros"]
    assert g.edge_labels == ["eu", "money", "us"]
    assert g.node_dict["e"].is_bus
    assert g.get_hierarchy_dict() == {"a": {"b": {}, "c": {}, "f": {}}, "d": {"e": {}}}
    assert g["a"].annotations.comment == "much nice, very comment"
    assert g["a", "b"][0].annotations.comment == "super useful"
    assert to_json(g) == path.read_text()


# JSON should default to default encoder for a random object and let that throw errors
def test_random_obj():
    class Dummy:
        pass

    with pytest.raises(TypeError):
        to_json(Dummy)


def test_from_json(datadir):
    with pytest.raises(ValueError) as e:
        from_json()
    assert str(e.value) == "`path` and `enc` arguments cannot both be `None`."

    with pytest.raises(ValueError) as e:
        from_json("", "")
    assert str(e.value) == "`path` and `enc` arguments cannot both be set."



"""Tests for _utils module."""

from copy import deepcopy

import pytest
from ragraph.analysis._utils import (
    clear_local_hierarchy,
    create_parent,
    get_available_name,
    inherit_edges,
    set_children,
    unset_parent,
)
from ragraph.graph import Graph, Node


def test_create_parent(graph: Graph):
    children = [graph.node_dict["a"], graph.node_dict["b"]]
    parent = create_parent(graph, children, prefix="cluster", lower=0)

    assert list(parent.children) == children, "Children should match."

    for child in children:
        assert child.parent == parent, "Parent should be set for children."

    assert parent in graph.nodes, "Parent should be added to the graph."


def test_get_available_name():
    g = Graph()

    name0 = get_available_name(g, prefix="test", lower=0)
    assert name0 == "test0"
    g.add_node(Node(name0))

    name1 = get_available_name(g, prefix="test", lower=0)
    assert name1 == "test1"
    g.add_node(Node(name1))  # Should fail if name already exists.


def test_inherit_edges(graph: Graph):
    ab = create_parent(graph, children=["a", "b"], name="ab")
    cd = create_parent(graph, children=["c", "d"], name="cd")
    inherit_edges(graph, [ab, cd], ["e", "f"])

    assert graph["ab", "e"], "Edge ab-e should exist."

    assert graph["e", "ab"], "Edge e-ab should exist."


def test_set_children_empty(graph: Graph):
    parent = Node("node")
    set_children(graph, parent, [])
    assert not set(parent.children), "Parent shouldn't have children."


def test_set_children_ab(graph: Graph):
    a = graph.node_dict["a"]
    b = graph.node_dict["b"]

    parent = Node("node")
    set_children(graph, parent, [a, b])
    assert set(parent.children) == {a, b}, "Children should be set to {a,b}."


def test_set_children_self(graph: Graph):
    parent = Node("node")
    set_children(graph, parent, [parent])
    assert not set(parent.children), "Should not have changed children."


def test_set_children_single_leaf(graph: Graph):
    parent = Node("node")
    e = graph.node_dict["e"]
    with pytest.raises(ValueError) as excinfo:
        set_children(graph, parent, [e])
        assert "Cannot set a single leaf node as children." in str(excinfo.value)


def test_set_children_single_parent(graph: Graph):
    a = graph.node_dict["a"]
    b = graph.node_dict["b"]
    p0 = create_parent(graph, [a, b], name="p0")
    p1 = Node("p1")
    set_children(graph, p1, [p0])


def test_clear_local_hierarchy():
    a = Node("a")
    b = Node("b")
    c = Node("c")
    c.parent = b
    b.parent = a
    g = Graph([a, b, c])
    clear_local_hierarchy(g, [c], [a])
    assert {a, c} == set(g.nodes), "'b' should no longer be in nodes and 'a' and 'c' should"


def test_unset_parent():
    """Unsetting the parent of "a" in the following tree

           h
       ____|____
      |         |
      f         g
     _|_     ___|___
    |   |   |   |   |
    a   b   c   d   e

    leaves the following forest:

    a      h
       ____|____
      |         |
      b         g
             ___|___
            |   |   |
            c   d   e

    in which "a" has become a root and "b" has become a child of "h" since "f" has
    been removed.

    etc.
    """
    a = Node("a")
    b = Node("b")
    c = Node("c")
    d = Node("d")
    e = Node("e")
    f = Node("f")
    g = Node("g")
    h = Node("h")

    h.children = [f, g]
    f.children = [a, b]
    g.children = [c, d, e]

    graph = Graph(nodes=[a, b, c, d, e, f, g, h])
    graph.check_consistency()
    hierarchy = graph.get_hierarchy_dict

    unset_parent(graph, h)  # Should do nothing
    assert hierarchy() == {"h": {"f": {"a": {}, "b": {}}, "g": {"c": {}, "d": {}, "e": {}}}}

    unset_parent(graph, a)
    assert hierarchy() == {"a": {}, "h": {"b": {}, "g": {"c": {}, "d": {}, "e": {}}}}

    unset_parent(graph, c)
    assert hierarchy() == {"a": {}, "c": {}, "h": {"b": {}, "g": {"d": {}, "e": {}}}}

    unset_parent(graph, d)
    assert hierarchy() == {"a": {}, "c": {}, "d": {}, "h": {"b": {}, "e": {}}}

    unset_parent(graph, e)
    assert hierarchy() == {"a": {}, "b": {}, "c": {}, "d": {}, "e": {}}


def test_unset_parent_and_clear(hierarchy):
    """Clearing a local hierarchy with protected root versus unsetting every leaf's
    parent should result in the same hierarchy."""
    g1 = deepcopy(hierarchy)
    g2 = deepcopy(hierarchy)

    clear_local_hierarchy(g1, g1.leafs, roots=list(g1.roots))

    for leaf in g2.leafs:
        unset_parent(g2, leaf, roots=list(g2.roots))

    assert g1.get_hierarchy_dict() == g2.get_hierarchy_dict()



"""Gamma bus detection module tests."""

from ragraph import datasets
from ragraph.analysis import bus
from ragraph.analysis._utils import create_parent
from ragraph.analysis.bus import gamma
from ragraph.graph import Edge, Graph


def test_bus_gamma_basic(parent_graph: Graph):
    bus_leafs, nonbus_leafs = gamma(
        parent_graph, root=parent_graph.roots[0].name, gamma=2.0, names=True
    )
    detected = set(bus_leafs)
    expected = {"e"}
    assert (
        detected == expected
    ), "{} should be the detected bus node for gamma 2.0, found {}.".format(expected, detected)

    bus_leafs, nonbus_leafs = gamma(
        parent_graph, root=parent_graph.roots[0].name, gamma=1.5, names=True
    )
    detected = set(bus_leafs)
    expected = {"a", "e"}
    assert (
        detected == expected
    ), "{} should be the detected bus nodes for gamma 1.5, found {}.".format(expected, detected)


def test_bus_gamma_no_bus_below_3_nodes(graph: Graph):
    bus_leafs, nonbus_leafs = gamma(graph, root="a", gamma=2.0)
    assert not bus_leafs, "There can never be bus nodes below 3 nodes."


def test_bus_gamma_no_bus(parent_graph: Graph):
    # Remove all edges.
    parent_graph.edges = []

    # No buses
    bus_leafs, nonbus_leafs = gamma(parent_graph, root=parent_graph.roots[0].name, gamma=2.0)
    assert not bus_leafs, "There can never be bus nodes if there are no edges."

    # "All bus nodes" -> no bus
    for i in parent_graph.node_list:
        for j in parent_graph.node_list:
            if i != j:
                parent_graph.add_edge(Edge(i, j))
    bus_leafs, nonbus_leafs = gamma(parent_graph, root=parent_graph.roots[0].name, gamma=1.0)
    assert not bus_leafs, "If all nodes are detected as bus, actually none should be."


def test_bus_gamma_climate_control():
    g = datasets.get("climate_control")
    r = create_parent(g, g.nodes)
    b, nb = bus.gamma(g, root=r.name, gamma=2.1, names=True)
    assert set(b) == {"Command Distribution"}


def test_bus_gamma_cases(case: Graph):
    r = create_parent(case, case.leafs)
    bus.gamma(case, root=r.name, gamma=2.0)



"""# Comparison analysis utilities"""

import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, FrozenSet, Tuple

from ragraph.edge import Edge
from ragraph.generic import Metadata
from ragraph.graph import Graph
from ragraph.node import Node

try:
    StrEnum = enum.StrEnum
except Exception:
    from strenum import StrEnum  # type: ignore
finally:
    if TYPE_CHECKING:
        StrEnum = enum.StrEnum


class NodeDescriptorLike(ABC):
    """A class that takes a node and describes it by returning a fixed (hashable) output.

    Hashable means that the description can be used to determine uniqueness by set operations.
    """

    @classmethod
    @abstractmethod
    def from_node(cls, node: Node) -> "NodeDescriptorLike":
        """Create a node descriptor from a node."""
        raise NotImplementedError

    @abstractmethod
    def to_node(self, graph: Graph) -> Node:
        """Create a node from a descriptor in a graph and return it for convenience."""
        raise NotImplementedError


@dataclass(frozen=True)
class NodeDescriptor(NodeDescriptorLike):
    """Describe a node based on its name, kind, labels and weights."""

    name: str
    kind: str
    labels: FrozenSet[str]
    weights: FrozenSet[Tuple[str, float]]

    @classmethod
    def from_node(cls, node: Node) -> "NodeDescriptor":
        return cls(
            node.name,
            node.kind,
            frozenset(node.labels),
            frozenset(node.weights.items()),
        )

    def to_node(self, graph: Graph) -> Node:
        node = Node(
            name=self.name,
            kind=self.kind,
            labels=list(self.labels),
            weights=dict(self.weights),
        )
        graph.add_node(node)
        return node


class EdgeDescriptorLike(ABC):
    """A class that takes an edge and describes it by returning a fixed (hashable) output.

    Hashable means that the description can be used to determine uniqueness by set operations.
    """

    @classmethod
    @abstractmethod
    def from_edge(cls, edge: Edge) -> "EdgeDescriptor":
        """Create an edge descriptor from an edge. Ignores annotations."""
        ...

    @abstractmethod
    def to_edge(self, graph: Graph) -> Edge:
        """Create an edge from a descriptor in a graph."""
        ...


@dataclass(frozen=True)
class EdgeDescriptor(EdgeDescriptorLike):
    """Describe a node based on its source name, target name, kind, labels, and weights."""

    source: str
    target: str
    kind: str
    labels: FrozenSet[str]
    weights: FrozenSet[Tuple[str, float]]

    @classmethod
    def from_edge(cls, edge: Edge) -> "EdgeDescriptor":
        return EdgeDescriptor(
            edge.source.name,
            edge.target.name,
            edge.kind,
            frozenset(edge.labels),
            frozenset(edge.weights.items()),
        )

    def to_edge(self, graph: Graph) -> Edge:
        edge = Edge(
            graph[self.source],
            graph[self.target],
            kind=self.kind,
            labels=list(self.labels),
            weights=dict(self.weights),
        )
        graph.add_edge(edge)
        return edge


class TagMode(StrEnum):
    """How to tag nodes as unique to one of the given inputs or common."""

    KIND = enum.auto()
    LABEL = enum.auto()
    ANNOTATION = enum.auto()


def tag(item: Metadata, tag: str, mode: TagMode):
    """Tag an item using a given tagging mode."""
    if mode == TagMode.KIND:
        item.kind = tag
    elif mode == TagMode.LABEL:
        item.labels = item.labels + [tag]
    elif mode == TagMode.ANNOTATION:
        item.annotations[tag] = True
    else:
        raise ValueError("Unknown tagging mode.")



"""pytest test configuration."""

import logging
from typing import List

import numpy as np
import pytest
from ragraph import datasets
from ragraph.graph import Edge, Graph, Node

# Toggle debugging output.
DEBUG_ANALYSIS = True


@pytest.fixture(autouse=DEBUG_ANALYSIS)
def analysis_logging():
    from ragraph.analysis import logger

    logger.setLevel(logging.DEBUG)
    yield logger
    logger.setLevel(logging.INFO)


@pytest.fixture(params=[i for i in datasets.enum() if i != "ledsip"])
def case(request):
    return datasets.get(request.param)


@pytest.fixture
def srcs():
    return [0, 1, 0, 2, 4, 4, 4, 4, 0, 1, 2, 3]


@pytest.fixture
def tgts():
    return [1, 0, 2, 0, 0, 1, 2, 3, 4, 4, 4, 4]


@pytest.fixture
def wgts():
    return [9, 9, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]


@pytest.fixture
def graph(srcs: List[int], tgts: List[int], wgts: List[int]):
    nodes = [Node(name) for name in ["a", "b", "c", "d", "e", "f"]]

    edges = [
        Edge(nodes[src], nodes[tgt], weights=dict(weight=wgt))
        for src, tgt, wgt in zip(srcs, tgts, wgts)
    ]

    return Graph(nodes=nodes, edges=edges)


@pytest.fixture
def parent_graph(graph: Graph):
    parent = Node("parent", children=graph.node_list)
    for node in graph.nodes:
        node.parent = parent
    graph.add_node(parent)
    return graph


@pytest.fixture
def matrix(srcs, tgts, wgts, graph):
    """Adjacency matrix that should correspond to the graph.
    [[0. 9. 1. 0. 1. 0.]
    [9. 0. 0. 0. 1. 0.]
    [1. 0. 0. 0. 1. 0.]
    [0. 0. 0. 0. 1. 0.]
    [1. 1. 1. 1. 0. 0.]
    [0. 0. 0. 0. 0. 0.]]
    """
    dim = graph.node_count
    matrix = np.zeros((dim, dim))
    for src, tgt, wgt in zip(srcs, tgts, wgts):
        matrix[src, tgt] = wgt
    return matrix


@pytest.fixture
def paper():
    Anb = np.array(
        [
            [0, 0, 1, 0, 1],
            [0, 0, 0, 0, 1],
            [0, 0, 0, 1, 0],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 0, 0],
        ]
    )
    nz = np.transpose(np.nonzero(Anb))
    nodes = [Node(n) for n in ["a", "b", "c", "d", "e"]]
    edges = [Edge(nodes[j], nodes[i]) for (i, j) in nz]
    return Graph(nodes=nodes, edges=edges)


@pytest.fixture
def chain_adj():
    """Simple chain of nodes."""
    return np.diag([1, 1, 1, 1], k=1)


@pytest.fixture
def chain_graph():
    """Simple chain of nodes."""
    ns = [Node(str(i)) for i in range(5)]
    es = [Edge(ns[i + 1], ns[i]) for i in range(4)]
    return Graph(nodes=ns, edges=es)


@pytest.fixture
def hierarchy():
    """Hierarchy with multiple levels and varying widths.

                       q
      _________________|_________________
      |                     |           |
      |              _______p______     |
      |              |            |     |
      |         _____o____        |     |
      |         |         |       |     |
    __k__   ____l___    __m__   __n__   |
    |   |   |   |   |   |   |   |   |   |
    a   b   c   d   e   f   g   h   i   j
    """
    a = Node("a")
    b = Node("b")
    c = Node("c")
    d = Node("d")
    e = Node("e")
    f = Node("f")
    g = Node("g")
    h = Node("h")
    i = Node("i")
    j = Node("j")
    k = Node("k")
    l = Node("l")  # noqa
    m = Node("m")
    n = Node("n")
    o = Node("o")
    p = Node("p")
    q = Node("q")

    q.children = [k, p, j]
    p.children = [o, n]
    o.children = [l, m]
    k.children = [a, b]
    l.children = [c, d, e]
    m.children = [f, g]
    n.children = [h, i]

    return Graph(nodes=[q], add_children=True)



"""Tests for generic.py module."""

import numpy as np
import pytest

from ragraph.datasets import esl
from ragraph.graph import ConsistencyError, Edge, Graph, Node


def test_empty_graph(empty_graph, a, b):
    assert empty_graph.node_count == 0, "empty_graph shouldn't have nodes."
    assert empty_graph.edge_count == 0, "empty_graph shouldn't have edges."
    assert list(empty_graph.edges_between(a, b)) == []
    assert list(empty_graph.edges_between_all([a], [b])) == []


def test_create_graph_ab(graph_ab):
    assert graph_ab.node_count == 2, "graph_ab should have 2 nodes."
    assert graph_ab.edge_count == 1, "graph_ab should have 1 edge."

    assert set(graph_ab.edges) == set(graph_ab.edges_from(graph_ab["a"]))

    assert set(graph_ab.edges) == set(graph_ab.edges_to(graph_ab["b"]))

    assert {graph_ab["b"]} == set(graph_ab.targets_of(graph_ab["a"]))

    assert {graph_ab["a"]} == set(graph_ab.sources_of(graph_ab["b"]))


def test_add_incorrect_node():
    with pytest.raises(TypeError):
        Graph(nodes=[1])

    with pytest.raises(ValueError):
        Graph(nodes=[Node("a"), Node("a")])


def test_add_incorrect_edge(graph_ab):
    c = Node("c")
    a = graph_ab["a"]
    with pytest.raises(ValueError) as e:
        graph_ab.add_edge(Edge(c, a))
    assert str(e.value) == "Source node {} does not exist in graph.".format(repr(c))

    with pytest.raises(ValueError) as e:
        graph_ab.add_edge(Edge(a, c))
    assert str(e.value) == "Target node {} does not exist in graph.".format(repr(c))


def test_add_dupe_edge(graph_ab):  # Has initial a->b edge.
    e = Edge(graph_ab["b"], graph_ab["a"])
    graph_ab.add_edge(e)
    graph_ab.add_edge(e)
    assert graph_ab.edge_count == 2, "Dupe edge should only be added once."


def test_delete_ab_nodes(graph_ab):
    a = graph_ab["a"]
    b = graph_ab["b"]
    graph_ab.del_node("a")
    graph_ab.del_node(b)
    assert graph_ab.nodes == []
    assert not graph_ab[a.name, b.name]
    assert graph_ab.max_depth == 0


def test_delete_ba_nodes(graph_ab):
    a = graph_ab["a"]
    b = graph_ab["b"]
    graph_ab.del_node(b)
    graph_ab.del_node("a")
    assert graph_ab.nodes == []
    assert not graph_ab[a.name, b.name]
    assert graph_ab.max_depth == 0


def test_delete_ab_edge(graph_ab):
    edge = graph_ab.edges[0]
    graph_ab.del_edge(edge)
    assert graph_ab.edges == []
    assert list(graph_ab.edges_from("a")) == []
    assert list(graph_ab.edges_to("b")) == []


def test_deletes_rich_a(rich_graph):
    rich_graph.del_node("a")
    assert rich_graph["c"].parent is None


def test_deletes_rich_c(rich_graph):
    c = rich_graph["c"]
    rich_graph.del_node(c)
    assert c not in rich_graph["a"].children


def test_rich_graph_properties(rich_graph):
    assert rich_graph.node_kinds == ["node", "rich", "wealthy"]
    assert rich_graph.edge_kinds == ["edge", "money"]
    assert rich_graph.get_nodes_by_kind("rich") == [rich_graph["c"]]
    assert rich_graph.get_edges_by_kind("money") == rich_graph["c", "d"]
    assert rich_graph.node_weight_labels == ["default", "dollars", "euros"]
    assert rich_graph.edge_weight_labels == ["default", "dollars"]
    assert rich_graph.edge_labels == ["default", "euros"]
    assert rich_graph.roots == rich_graph.node_list[:2]  # a, b
    assert rich_graph.leafs == rich_graph.node_list[1:]  # b, c, d
    assert rich_graph.max_depth == 1
    assert list(rich_graph.edges_between_all(["c"], ["d"])) == rich_graph["c", "d"]
    assert rich_graph.get_hierarchy_dict() == {"a": {"c": {}, "d": {}}, "b": {}}
    assert rich_graph.check_consistency(raise_error=False)
    assert rich_graph.get_hierarchy_dict(levels=0) == {"a": {}, "b": {}}


def test_get_graph_slice(rich_graph):
    gslice = rich_graph.get_graph_slice(nodes=rich_graph.nodes[:2])
    assert gslice.node_count == 2
    assert gslice.edge_count == 1

    gslice = rich_graph.get_graph_slice(nodes=[rich_graph["a"], rich_graph["c"]])
    assert gslice["c"].parent == gslice["a"]


def test_no_edges_from(graph_ab):
    assert list(graph_ab.edges_from(graph_ab["b"])) == [], "Should be no edges from 'b'."


def test_inconsistency_recursion(graph_ab):
    a = graph_ab["a"]
    b = graph_ab["b"]
    a.parent = b
    b.parent = a
    with pytest.raises(ConsistencyError) as e:
        graph_ab.check_consistency()
    assert str(e.value) == "Node {} is in its own ancestors.".format(a)

    a._parent = None
    with pytest.raises(ConsistencyError) as e:
        graph_ab.check_consistency()
    assert str(e.value) == "Node {} is in its own descendants.".format(a)


def test_inconsistency_parent(graph_ab, c):
    a = graph_ab["a"]
    b = graph_ab["b"]

    a._parent = c
    with pytest.raises(ConsistencyError) as e:
        graph_ab.check_consistency()
    assert str(e.value) == "Node {}'s parent {} is missing in the graph.".format(a, c)

    c._children = []
    graph_ab.add_node(c)
    with pytest.raises(ConsistencyError) as e:
        graph_ab.check_consistency()
    assert str(e.value) == "Node {}'s does not occur in parent {}'s children.".format(a, c)

    c._children = [a, b]
    b._parent = None
    with pytest.raises(ConsistencyError) as e:
        graph_ab.check_consistency()
    assert str(e.value) == "Node {}'s child has a different parent {}.".format(c, b.parent)
    b._parent = c

    c._is_bus = True
    with pytest.raises(ConsistencyError) as e:
        graph_ab.check_consistency()
    assert str(e.value) == "Node {} is a bus node, but has no parent to be it for.".format(c)

    graph_ab._nodes.pop("a")
    with pytest.raises(ConsistencyError) as e:
        graph_ab.check_consistency()
    assert str(e.value) == "Node {}'s child {} is missing in the graph.".format(c, a)


def test_art_empty_graph(empty_graph):
    with pytest.raises(ValueError) as e:
        empty_graph.get_ascii_art()
    assert str(e.value) == "Empty graph, cannot create ASCII art."


def test_art_graph_ab(graph_ab, capsys):
    art = graph_ab.get_ascii_art(show=False)
    ref = """ ┌───┬───┐
a┥ ■ │   │
 ├───┼───┤
b┥ X │ ■ │
 └───┴───┘"""
    assert art == ref

    graph_ab.get_ascii_art(show=True)
    captured = capsys.readouterr()
    assert captured.out == art + "\n"

    a = graph_ab["a"]
    art = graph_ab.get_ascii_art(nodes=[a], show=False)
    ref = """ ┌───┐
a┥ ■ │
 └───┘"""
    assert art == ref


def test_adj_graph_ab(graph_ab):
    adj = graph_ab.get_adjacency_matrix()

    mat = [[0.0, 0.0], [1.0, 0.0]]
    if np:
        assert (adj == np.array(mat)).all()
    else:
        assert adj == mat


def test_matrices_rich_graph(rich_graph: Graph):
    b = rich_graph["b"]
    e = Node("e", parent=b)
    rich_graph.add_node(e)
    c = rich_graph["c"]  # Child of a.

    # Add an edge between children of a and b
    rich_graph.add_edge(Edge(c, e, weights=dict(default=5)))

    # Add a self loop to b
    rich_graph.add_edge(Edge(b, b, weights=dict(default=3)))

    gam = rich_graph.get_adjacency_matrix
    roots = rich_graph.roots

    assert (gam(roots, inherit=False, loops=False) == np.array([[0.0, 0.0], [1.0, 0.0]])).all()
    assert (gam(roots, inherit=True, loops=False) == np.array([[0.0, 0.0], [6.0, 0.0]])).all()
    assert (gam(roots, inherit=True, loops=True) == np.array([[1.0, 0.0], [6.0, 3.0]])).all()
    assert (gam(roots, inherit=False, loops=True) == np.array([[0.0, 0.0], [1.0, 3.0]])).all()
    assert (
        gam(roots, inherit=True, loops=True, only=["dollars"]) == np.array([[1.0, 0.0], [0.0, 0.0]])
    ).all()
    gmm = rich_graph.get_mapping_matrix
    assert np.all(gmm(roots, roots) == gam(roots))
    assert (gmm(rows=["e"], cols=["c", "d"]) == np.array([[5.0, 0.0]])).all()


def test_graph_indexing(rich_graph):
    assert rich_graph["a"] == rich_graph["a"]
    assert rich_graph["a", "b"] == rich_graph["a", "b"]
    with pytest.raises(TypeError) as e:
        rich_graph[{"q"}]
    assert "should be a node name" in str(e.value)


def test_graph_contains(rich_graph):
    assert "a" in rich_graph
    assert ("a", "b") in rich_graph
    assert "dummy" not in rich_graph
    with pytest.raises(TypeError) as e:
        {"q"} in rich_graph
    assert "should be a node name" in str(e.value)


def test_del_node_inherit(a, b, c, d, e, f):
    a.children = [b, c, d]
    d.children = [e, f]
    g = Graph(nodes=[a, b, c, d, e, f])
    g.del_node(d, inherit=True)
    assert a.children == [b, c, e, f]


def test_graph_get_node_and_edge_selection():
    g = esl.get("pump")

    # Dependent selection
    nodes, edges = g.get_node_and_edge_selection(
        node_kinds=["component", "function_spec", "variable"],
        edge_kinds=["functional_dependency", "mapping_dependency"],
        depth=2,
        selection_mode="dependent",
    )

    # Check selected nodes.
    assert sorted([node.name for node in nodes]) == [
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor-control-signal",
        "world.drive-mechanism.motor.convert-power",
        "world.drive-mechanism.power",
        "world.drive-mechanism.power-potential",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-source.convert-potential",
        "world.drive-mechanism.power-switch",
        "world.drive-mechanism.provide-power",
        "world.drive-mechanism.send-control-signal",
        "world.provide-torque",
        "world.pump",
        "world.pump-length",
        "world.pump.convert-torque",
        "world.torque",
        "world.water-flow",
    ]

    # Check sources of selected edges of kind functional_dependency.
    assert sorted([edge.source.name for edge in edges if edge.kind == "functional_dependency"]) == [
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor.convert-power",
        "world.drive-mechanism.power",
        "world.drive-mechanism.power-potential",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-source.convert-potential",
        "world.drive-mechanism.power-switch",
        "world.drive-mechanism.provide-power",
        "world.provide-torque",
        "world.pump",
        "world.pump",
        "world.pump",
        "world.torque",
    ]

    # Check targets of selected edges of kind functional_dependency
    assert sorted([edge.target.name for edge in edges if edge.kind == "functional_dependency"]) == [
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor.convert-power",
        "world.drive-mechanism.power",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-switch",
        "world.drive-mechanism.provide-power",
        "world.provide-torque",
        "world.pump",
        "world.pump",
        "world.pump",
        "world.pump.convert-torque",
        "world.torque",
        "world.water-flow",
    ]

    # Check sources of selected edges of kind mapping_dependency.
    assert sorted([edge.source.name for edge in edges if edge.kind == "mapping_dependency"]) == [
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor",
        "world.drive-mechanism.motor.convert-power",
        "world.drive-mechanism.motor.convert-power",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-source",
        "world.drive-mechanism.power-source.convert-potential",
        "world.drive-mechanism.power-source.convert-potential",
        "world.drive-mechanism.power-switch",
        "world.drive-mechanism.power-switch",
        "world.drive-mechanism.provide-power",
        "world.drive-mechanism.send-control-signal",
        "world.provide-torque",
        "world.pump",
        "world.pump",
        "world.pump",
        "world.pump",
        "world.pump",
        "world.pump",
        "world.pump.convert-torque",
        "world.pump.convert-torque",
    ]

    # Check targets of selected edges of kind mapping_dependency
    assert sorted([edge.target.name for edge in edges if edge.kind == "mapping_dependency"]) == [
        "world.drive-mechanism.motor-control-signal",
        "world.drive-mechanism.motor-control-signal",
        "world.drive-mechanism.motor-control-signal",
        "world.drive-mechanism.motor.convert-power",
        "world.drive-mechanism.power",
        "world.drive-mechanism.power",
        "world.drive-mechanism.power",
        "world.drive-mechanism.power",
        "world.drive-mechanism.power",
        "world.drive-mechanism.power",
        "world.drive-mechanism.power",
        "world.drive-mechanism.power-potential",
        "world.drive-mechanism.power-potential",
        "world.drive-mechanism.power-source.convert-potential",
        "world.drive-mechanism.provide-power",
        "world.drive-mechanism.provide-power",
        "world.drive-mechanism.send-control-signal",
        "world.drive-mechanism.send-control-signal",
        "world.provide-torque",
        "world.provide-torque",
        "world.pump-length",
        "world.pump.convert-torque",
        "world.torque",
        "world.torque",
        "world.torque",
        "world.torque",
        "world.torque",
        "world.torque",
        "world.torque",
        "world.water-flow",
        "world.water-flow",
    ]

    # Check if no other edge kinds are selected.
    assert set([e.kind for e in edges]) == {
        "functional_dependency",
        "mapping_dependency",
    }

    # Check selected nodes based on different depth and selection mode.
    nodes, edges = g.get_node_and_edge_selection(
        node_kinds=["component", "function_spec", "variable"],
        edge_kinds=["functional_dependency", "mapping_dependency"],
        depth=1,
        selection_mode="independent",
    )

    assert sorted([node.name for node in nodes]) == [
        "world.drive-length",
        "world.drive-mechanism",
        "world.drive-mechanism.convert-power-potential",
        "world.drive-mechanism.motor-control-signal",
        "world.drive-mechanism.motor.conversion",
        "world.drive-mechanism.motor.convert-power",
        "world.drive-mechanism.power",
        "world.drive-mechanism.power-potential",
        "world.drive-mechanism.power-source.convert-potential",
        "world.drive-mechanism.provide-power",
        "world.drive-mechanism.send-control-signal",
        "world.provide-torque",
        "world.pump",
        "world.pump-length",
        "world.pump.convert-torque",
        "world.torque",
        "world.water-flow",
    ]



"""Tests for the plot colors module."""

import pytest
from ragraph import colors
from ragraph.colors import cubehelix, utils


@pytest.fixture(
    params=[
        (
            "get_categorical",
            [
                "#6a9ffc",
                "#dc69e2",
                "#f27663",
                "#8db32f",
                "#38cc8e",
            ],
        ),
        (
            "get_continuous",
            [
                "#b6ead9",
                "#7dc2d5",
                "#6b86cd",
                "#6b499f",
                "#541d4f",
            ],
        ),
        (
            "get_hue",
            [
                "#ffffff",
                "#ffb9ad",
                "#f67d6a",
                "#cb4b38",
                "#852415",
            ],
        ),
        (
            "get_purple",
            [
                "#ffffff",
                "#debcff",
                "#b581ff",
                "#8750da",
                "#512890",
            ],
        ),
        (
            "get_red",
            [
                "#ffffff",
                "#ffb6b3",
                "#fa7974",
                "#cf4842",
                "#88211d",
            ],
        ),
        (
            "get_orange",
            [
                "#ffffff",
                "#f8c597",
                "#de8f4a",
                "#b25f15",
                "#723300",
            ],
        ),
        (
            "get_green",
            [
                "#ffffff",
                "#afea97",
                "#6dc94a",
                "#3b9b15",
                "#186100",
            ],
        ),
        (
            "get_cyan",
            [
                "#ffffff",
                "#94e8eb",
                "#45c6cb",
                "#10989d",
                "#005e62",
            ],
        ),
        (
            "get_blue",
            [
                "#ffffff",
                "#b7cfff",
                "#7a9fff",
                "#4870da",
                "#224090",
            ],
        ),
        (
            "get_diverging_hues",
            [
                "#88211d",
                "#fa7974",
                "#ffffff",
                "#7a9fff",
                "#224090",
            ],
        ),
        (
            "get_diverging_redblue",
            ["#88211d", "#fa7974", "#ffffff", "#7a9fff", "#224090"],
        ),
        (
            "get_diverging_orangecyan",
            ["#723300", "#de8f4a", "#ffffff", "#45c6cb", "#005e62"],
        ),
        (
            "get_diverging_purplegreen",
            ["#512890", "#b581ff", "#ffffff", "#6dc94a", "#186100"],
        ),
    ]
)
def palettes_exp(request):
    fname, exp5 = request.param
    fn = getattr(colors, fname)
    return fn, exp5


def test_toplevel_palettes(palettes_exp):
    # Unpack the fixture.
    fn, exp5 = palettes_exp

    N = 5  # uneven as we'd like to test middle points in the diverging pals.
    result = fn(N)

    # Expected result for N=5.
    assert len(result) == 5
    assert result == exp5

    # Some smoke testing.
    assert len(fn(1)) == 1
    assert len(fn(0)) == 1


def test_categorical_extra():
    result = colors.get_categorical(12)
    assert result == [
        "#4e85e7",
        "#de4eb2",
        "#ba8322",
        "#35c360",
        "#947cff",
        "#fa668b",
        "#99b030",
        "#40cfab",
        "#d57cf8",
        "#f88c6e",
        "#80d45e",
        "#69ceea",
    ]


def test_cubehelix():
    pure = [
        [round(i, 10) for i in color]
        for color in cubehelix._cubehelix_pure_python(5, categorical=True)
    ]
    nump = [
        [round(i, 10) for i in color] for color in cubehelix._cubehelix_numpy(5, categorical=True)
    ]

    assert pure == nump, "Should be identical except from rounding errors."

    # Fully cover the cubehelix utils, too.
    shuffle_me = [0, 1, 2, 3]
    assert cubehelix._shuffle(shuffle_me, 2) == [0, 2, 1, 3], "should return shuffled."
    assert cubehelix._shuffle(shuffle_me, 8) == [0, 1, 2, 3], "should return input."


def test_utils():
    assert utils.rgb_to_rgb1(51, 0, 255) == (0.2, 0.0, 1.0)
    assert utils.hex_to_rgb("#bffbae") == (191, 251, 174)



"""Tests for Johnson's circuit finding algorithm."""

from ragraph.analysis import heuristics


def test_heuristic_johnson_cc(graph):
    circuits = list(heuristics.johnson(graph))
    circuits = {tuple(n.name for n in c) for c in circuits}
    ref = {
        ("a", "b", "e", "c"),
        ("a", "c", "e", "b"),
        ("a", "b", "e"),
        ("a", "c", "e"),
        ("a", "e", "b"),
        ("a", "e", "c"),
        ("a", "b"),
        ("a", "c"),
        ("a", "e"),
        ("b", "e"),
        ("c", "e"),
        ("d", "e"),
    }
    assert circuits == ref



import pytest
from ragraph.io import matrix


def test_from_matrix():
    mat = [[1, 0, 1, 0], [0, 2, 3, 4]]
    rows = ["a", "b"]
    cols = ["a", "c", "d", "e"]
    weight_label = "num"
    empty = 0
    graph = matrix.from_matrix(
        mat, rows, cols=cols, weight_label=weight_label, empty=empty, name="graph_ab"
    )
    assert set(graph.node_dict.keys()) == {"a", "b", "c", "d", "e"}
    assert len(graph.edges) == 5
    assert graph.edge_weight_labels == ["num"]


def test_round_trip(graph_ab):
    mat = matrix.to_matrix(graph_ab)
    assert graph_ab == matrix.from_matrix(mat, graph_ab.nodes, name="graph_ab")
    assert matrix.from_matrix(mat).node_dict.get("node0", False)


def test_errors():
    with pytest.raises(ValueError) as e:
        matrix.from_matrix([[0]], ["a", "b"])
        assert "matrix dimensions" in str(e.value).lower()


def test_round_trip_no_numpy(graph_ab):
    matrix.np = None

    mat = matrix.to_matrix(graph_ab)
    assert graph_ab == matrix.from_matrix(mat, graph_ab.nodes, name="graph_ab")

    try:
        import numpy as np

        matrix.np = np
    except ImportError:
        pass



from ragraph import datasets
from ragraph.analysis import cluster, sequence
from ragraph.graph import Graph
from ragraph.node import Node


def test_branchsort():
    g = datasets.get("tarjans8")

    g.node_dict["1"].children = [g.node_dict[n] for n in ["7", "4", "5"]]
    g.node_dict["2"].children = [g.node_dict[n] for n in ["8", "3", "6"]]

    roots = ["2", "1"]

    graph, roots, leafs = sequence.utils.branchsort(
        algo=sequence.name,
        graph=g,
        root=None,
        nodes=roots,
        inplace=True,
        recurse=True,
        names=True,
    )

    s1 = ["4", "5", "7"]
    s2 = ["3", "6", "8"]

    assert roots == ["1", "2"]
    assert [n.name for n in graph.node_dict[roots[0]].children] == s1
    assert [n.name for n in graph.node_dict[roots[1]].children] == s2
    assert leafs == s1 + s2


def test_branchsort_cases(case):
    graph, roots = cluster.hierarchical_markov(case, inplace=True)
    graph, roots, leafs = sequence.utils.branchsort(
        algo=sequence.markov,
        graph=case,
        root=None,
        recurse=False,
        inplace=True,
        algo_args=dict(inf=1.0, dep=1.0, mu=2.0),
    )


def test_branchsort_leafs():
    """Test whether the branchsort util keeps leaves intact."""
    nodes = {i: Node(name=i) for i in "abcdefghijkl"}
    nodes["a"].children = [nodes["c"], nodes["b"]]  # notice it's reversed.
    nodes["d"].children = [nodes["f"], nodes["e"]]  # also.
    nodes["f"].children = [nodes["h"], nodes["g"]]  # also.
    nodes["e"].children = [nodes["j"], nodes["i"]]  # also.
    nodes["h"].children = [nodes["l"], nodes["k"]]  # also.

    graph = Graph(nodes=nodes.values())

    # Sequence by name, but set "a" and "f" as leaf nodes (e.g. protect them).
    graph, roots, leafs = sequence.utils.branchsort(
        algo=sequence.name,
        graph=graph,
        root=None,
        recurse=True,
        inplace=True,
        nodes=graph.roots,
        leafs=["a", "f"],
    )

    assert nodes["a"].children == [nodes["c"], nodes["b"]], "Should be intact."
    assert nodes["d"].children == [nodes["f"], nodes["e"]][::-1], "Should be reversed."
    assert nodes["f"].children == [nodes["h"], nodes["g"]], "Should be intact."
    assert nodes["e"].children == [nodes["j"], nodes["i"]][::-1], "Should be reversed."
    assert nodes["h"].children == [nodes["l"], nodes["k"]], "Should be intact."



from pytest import mark
from ragraph import datasets
from ragraph.plot import dmm
from ragraph.plot.generic import Style


@mark.skip(reason="Dict comparision fails randomly.")
def test_dmm_0(check_plotly):
    g = datasets.get("climate_control")
    rows = g.leafs[0:5]
    cols = g.leafs
    fig = dmm(
        rows=rows,
        cols=cols,
        edges=[e for e in g.edges_between_all(sources=cols, targets=rows)],
    )

    fname = "dmm_0.json"
    check_plotly(fig, fname)


@mark.skip(reason="Dict comparision fails randomly.")
def test_dmm_1(check_plotly):
    g = datasets.get("climate_control")

    # Categorical plot of weight labels with custom colors
    rows = g.leafs
    cols = g.leafs[0:5]
    fig = dmm(
        rows=rows,
        cols=cols,
        edges=[e for e in g.edges_between_all(sources=cols, targets=rows)],
    )

    fname = "dmm_1.json"
    check_plotly(fig, fname)


@mark.skip(reason="Dict comparision fails randomly.")
def test_dmm_2(check_plotly):
    g = datasets.get("climate_control")

    # Categorical plot of weight labels with custom colors
    rows = g.leafs
    cols = g.leafs[0:5]
    fig = dmm(
        rows=rows,
        cols=cols,
        edges=[e for e in g.edges_between_all(sources=cols, targets=rows)],
        style=Style(row_col_numbers=False),
    )

    fname = "dmm_2.json"
    check_plotly(fig, fname)
